# -*- coding: utf-8 -*-
"""
Created on Tue Oct 25 10:50:54 2022

@author: ereznic2
"""
import os
import pandas as pd
import numpy as np
import time
import subprocess

def run_RODeO(atb_year,site_name,turbine_model,electrolysis_scale,policy_option,policy,i,wind_size_mw,solar_size_mw,electrolyzer_size_mw,\
              energy_to_electrolyzer,electrolyzer_energy_kWh_per_kg,hybrid_plant,revised_renewable_cost,electrolyzer_capex_kw,capex_ratio_dist,wind_om_cost_kw,\
              useful_life,time_between_replacement,\
              grid_connection_scenario,grid_price_scenario,gams_locations_rodeo_version,rodeo_output_dir):

     # Renewable generation profile
     system_rating_mw = wind_size_mw + solar_size_mw
     # Renewable output profile needs to be same length as number of time periods in RODeO.
     # Ideally it would be 8760 but if for some reason a couple hours less, this is a simple fix
     while len(energy_to_electrolyzer)<8760:
         energy_to_electrolyzer.append(energy_to_electrolyzer[-1])
         
     electrical_generation_timeseries = np.zeros_like(energy_to_electrolyzer)
     electrical_generation_timeseries[:] = energy_to_electrolyzer[:]
     # Put electrolyzer input into MW
     electrical_generation_timeseries = electrical_generation_timeseries/1000
     # Normalize renewable profile to system rating (RODeO then scales it back up)
     electrical_generation_timeseries = electrical_generation_timeseries/system_rating_mw
     # Get renewable generation profile into a format that works for RODeO
     electrical_generation_timeseries_df = pd.DataFrame(electrical_generation_timeseries).reset_index().rename(columns = {'index':'Interval',0:1})
     electrical_generation_timeseries_df['Interval'] = electrical_generation_timeseries_df['Interval']+1
     electrical_generation_timeseries_df = electrical_generation_timeseries_df.set_index('Interval')
     
     # Fill in renewable profile for RODeO with zeros for years 2-20 (because for some reason it neesd this)
     extra_zeroes = np.zeros_like(energy_to_electrolyzer)
     for j in range(19):
         #j=0
         extra_zeroes_df = pd.DataFrame(extra_zeroes,columns = [j+2]).reset_index().rename(columns = {'index':'Interval',0:j+2})
         extra_zeroes_df['Interval'] = extra_zeroes_df['Interval']+1
         extra_zeroes_df = extra_zeroes_df.set_index('Interval')
         electrical_generation_timeseries_df = electrical_generation_timeseries_df.join(extra_zeroes_df)
         # normalized_demand_df = normalized_demand_df.join(extra_zeroes_df)

     # Write the renewable generation profile to a .csv file in the RODeO repository, assuming RODeO is installed in the same folder as HOPP
     ren_profile_name = 'ren_profile_'+str(atb_year) + '_'+site_name + '_'+ turbine_model
     electrical_generation_timeseries_df.to_csv("examples/H2_Analysis/RODeO_files/Data_files/TXT_files/Ren_profile/" + ren_profile_name + '.csv',sep = ',')
     
     equation_year_CEPCI = 603.1
     model_year_CEPCI = 607.5
     
     # Storage costs as a function of location. Eventuall change this to specify type of storage to RODeO
     if site_name == 'IA':
         h2_storage_cost_USDperkg = 540
         #balancing_area = 'p65'
     elif site_name == 'TX':
         h2_storage_cost_USDperkg = model_year_CEPCI/equation_year_CEPCI*12.30
         #balancing_area ='p124'
     elif site_name == 'IN':
         h2_storage_cost_USDperkg = 540
         #balancing_area = 'p128'
     elif site_name == 'MS':
         h2_storage_cost_USDperkg = model_year_CEPCI/equation_year_CEPCI*12.30
         #balancing_area = 'p9'
         
     # Municipal water rates and wastewater treatment rates combined ($/gal)
     if site_name == 'IA':
         water_cost = 0.00612
         #balancing_area = 'p65'
     elif site_name == 'TX':
         water_cost = 0.00811
         #balancing_area ='p124'
     elif site_name == 'IN':
         water_cost = 0.00634
         #balancing_area = 'p128'
     elif site_name == 'MS':
         water_cost = 0.00844  
         #balancing_area = 'p9'

     
     # Format renewable system cost for RODeO
     #hybrid_installed_cost = hybrid_plant.grid.total_installed_cost
     hybrid_installed_cost = revised_renewable_cost
     hybrid_installed_cost_perMW = hybrid_installed_cost/system_rating_mw  
     
     
     #if electrolysis_scale == 'Centralized':
         # Installed capital cost
     electrolyzer_installation_factor = 12/100  #[%] for stack cost 
    
    # Indirect capital cost as a percentage of installed capital cost
     site_prep = 2/100   #[%]
     engineering_design = 10/100 #[%]
     permitting = 15/100     #[%]         
     project_contingency = 15/100 #[%]
    #[%]
     land_cost = 250000   #[$]
     # elif electrolysis_scale == 'Distributed':
     #     electrolyzer_installation_factor = 12/100/capex_ratio_dist  #[%] for stack cost 
         
     #     # Indirect capital cost as a percentage of installed capital cost
     #     site_prep = 0/100   #[%]
     #     engineering_design = 0/100 #[%]
     #     permitting = 15/100/capex_ratio_dist     #[%]         
     #     project_contingency = 15/100/capex_ratio_dist #[%]
     #     land_cost = 0   #[$]
     

     stack_replacement_cost = 15/100  #[% of installed capital cost]
     fixed_OM = 0.24     #[$/kg H2]    
     
     # Calculate electrolyzer installation cost
     total_direct_electrolyzer_cost_kw = (electrolyzer_capex_kw * (1+electrolyzer_installation_factor)) \
     
     electrolyzer_total_installed_capex = total_direct_electrolyzer_cost_kw*electrolyzer_size_mw*1000
     
     electrolyzer_indirect_cost = electrolyzer_total_installed_capex*(site_prep+engineering_design+project_contingency+permitting)
     
     #electrolyzer_installation_cost = electrolyzer_system_capex_kw*stack_installation_factor*electrolyzer_size_mw\
     #                               + electrolyzer_indirect_cost                             
     
     compressor_capex_USDprkWe_of_electrolysis = 39
     
     # Calculate capital costs
     electrolyzer_total_capital_cost = electrolyzer_total_installed_capex + electrolyzer_indirect_cost
         
     electrolyzer_system_capex_kw = electrolyzer_total_capital_cost/electrolyzer_size_mw/1000
     
     # O&M costs
     # https://www.sciencedirect.com/science/article/pii/S2542435121003068
     fixed_OM = 12.8 #[$/kWh-y]
     property_tax_insurance = 1.5/100    #[% of Cap/y]
     variable_OM = 1.30  #[$/MWh]
     
          
     # Tell RODeO if grid connected or not
     if grid_connection_scenario == 'off-grid':
         grid_connected_rodeo = False
         grid_string = grid_connection_scenario
         grid_imports = 0
         # Electrolyzer CF estimation for ammortized variable O&M cost calculation
         elec_cf = sum(energy_to_electrolyzer)/(electrolyzer_size_mw*1000*8760)
     elif grid_connection_scenario == 'grid-only':
         grid_connected_rodeo = True
         grid_string = grid_connection_scenario+'-'+grid_price_scenario
         grid_imports = 1
         # Electrolyzer CF estimation for ammortized variable O&M cost calculation
         elec_cf = 0.97
     elif grid_connection_scenario == 'hybrid-grid':
         grid_connected_rodeo = True
         grid_string = grid_connection_scenario+'-'+grid_price_scenario
         grid_imports = 1
         # Electrolyzer CF estimation for ammortized variable O&M cost calculation
         elec_cf = 0.97

     # Amortized refurbishment expense [$/MWh]
     amortized_refurbish_cost = (total_direct_electrolyzer_cost_kw*stack_replacement_cost)\
             *max(((useful_life*8760*elec_cf)/time_between_replacement-1),0)/useful_life/8760/elec_cf*1000

     total_variable_OM = variable_OM+amortized_refurbish_cost
     
     # Define electrolyzer capex, fixed opex, and energy consumption (if not pulling from external data)
     electrolyzer_capex_USD_per_MW = electrolyzer_system_capex_kw*1000
     electrolyzer_fixed_opex_USD_per_MW_year = fixed_OM*1000
     #electrolyzer_energy_kWh_per_kg = 55.5 # Eventually get from input loop
     
     # Define dealination conversion factors
     desal_energy_conversion_factor_kWh_per_m3_water = 4 # kWh per m3-H2O
     m3_water_per_kg_h2 = 0.01 # m3-H2O per kg-H2
     
     # Calculate desalination energy requirement per kg of produced hydrogen
     desal_energy_kWh_per_kg_H2 = m3_water_per_kg_h2*desal_energy_conversion_factor_kWh_per_m3_water
     
     # Desalination system size
     desal_sys_size_m3_hr = electrolyzer_size_mw*1000/electrolyzer_energy_kWh_per_kg*m3_water_per_kg_h2
     desal_sys_size_mw = desal_sys_size_m3_hr*desal_energy_conversion_factor_kWh_per_m3_water/1000
     
     # Total desal capex and opex
     desal_capex_total = 32894 * (997 * desal_sys_size_m3_hr / 3600) # Output in USD
     desal_opex_total = 4841 * (997 * desal_sys_size_m3_hr / 3600) # Output in USD/yr
     
     # Desal capex and opex in $/MW
     desal_capex_per_mw = desal_capex_total/desal_sys_size_mw
     desal_opex_per_mw = desal_opex_total/desal_sys_size_mw
     
     # Calculate desal capex and opex per MW of electrolysis power
     # desal_capex_USD_per_MW_of_electrolysis = 32894*(997/3600*1000/electrolyzer_energy_kWh_per_kg*m3_water_per_kg_h2)
     # desal_opex_USD_per_MW_of_EC_per_year = 4841*(997/3600*1000/electrolyzer_energy_kWh_per_kg*m3_water_per_kg_h2)
     
     # Incorporate desal cost and efficiency into electrolyzer capex, opex, and energy consumption
     #electrolysis_desal_total_capex_per_MW = electrolyzer_capex_USD_per_MW + desal_capex_USD_per_MW_of_electrolysis
     #electrolysis_desal_total_opex_per_MW_per_year = electrolyzer_fixed_opex_USD_per_MW_year + desal_opex_USD_per_MW_of_EC_per_year
     electrolysis_desal_total_energy_consumption = electrolyzer_energy_kWh_per_kg + desal_energy_kWh_per_kg_H2
     
     # Convert electrolysis energy consumption into LHV efficiency
     hydrogen_LHV = 120000 #kJ/kg
     eta_LHV = hydrogen_LHV/3600/electrolysis_desal_total_energy_consumption

     
         
     # Financial parameters
     inflation_rate = 0.0/100
     equity_percentage = 42/100
     bonus_depreciation = 0/100
     
     # Set hydrogen break even price guess value
     # Could in the future replace with H2OPP or H2A estimates 
     lcoh_guessvalue =50
     # Placeholder for if not doing optimization; may want to move this elsewhere or higher level
     h2_storage_duration = 10
     optimize_storage_duration = 1
     
     # Specify grid cost year for ATB year
     if atb_year == 2020:
         grid_year = 2025
     elif atb_year == 2025:
         grid_year = 2030
     elif atb_year == 2030:
         grid_year = 2035
     elif atb_year == 2035:
         grid_year = 2040
         
     # Policy impacts
     policy_scenario = policy[i]
        
     # Set up batch file
     dir0 = "..\\RODeO\\"
     dir1 = 'examples\\H2_Analysis\\RODeO_files\\Data_files\\TXT_files\\'
     dirout = rodeo_output_dir
     
    # txt1 = '"C:\\GAMS\\win64\\24.8\\gams.exe" ..\\RODeO\\Storage_dispatch_SCS license=C:\\GAMS\\win64\\24.8\\gamslice.txt'
     txt1 = gams_locations_rodeo_version[0]
     
     # # Putting this as a conditional just so it will run with existing data, but we should change this to the second one eventually
     # if electrolysis_scale == 'Centralized':
     #     scenario_name = str(atb_year)+'_'+ site_name +'_'+turbine_model+'_'+grid_string
     # elif electrolysis_scale == 'Distributed':
     #     scenario_name = str(atb_year)+'_'+ site_name +'_'+turbine_model+'_'+policy_option.replace(' ','-') + '_'+electrolysis_scale+ '_' + grid_string
         
     scenario_name = str(atb_year)+'_'+ site_name +'_'+turbine_model+'_'+policy_option.replace(' ','-') + '_'+electrolysis_scale+ '_' + grid_string
     
     scenario_inst = ' --file_name_instance='+scenario_name
     #scenario_name = ' --file_name_instance='+Scenario1
     # demand_prof = ' --product_consumed_inst=' + dem_profile_name
     demand_prof = ' --product_consumed_inst=Product_consumption_flat_hourly_ones'
     load_prof = ' --load_prof_instance=Additional_load_none_hourly'
     ren_prof = ' --ren_prof_instance=Ren_profile\\'+ren_profile_name

     
     #energy_price = ' --energy_purchase_price_inst=Elec_prices\\Elec_purch_price_wholesale_MWh_hourly'
     energy_price = ' --energy_purchase_price_inst=Elec_prices\\Elec_purch_price_MWh_MC95by35_'+grid_price_scenario+'_'+site_name+'_'+str(grid_year)

     #energy_price = ' --energy_purchase_price_inst=Elec_prices\\Elec_purch_price_WS_MWh_MC95by35_'+str(balancing_area)+'_'+str(atb_year)
     #energy_price = ' --energy_purchase_price_inst=Netload_'+str(i1)+' --energy_sale_price_inst=Netload_'+str(i1)
     #max_input_entry = ' --Max_input_prof_inst=Max_input_cap_'+str(i1)
     #capacity_values = ' --input_cap_instance='+str(electrolyzer_size_mw)#+str(storage_power_increment)#+' --output_cap_instance='+str(storage_power_increment)
     efficiency = ' --input_efficiency_inst='+str(round(eta_LHV,4))#'0.611'#+str(round(math.sqrt(RTE[i1-1]),6))#+' --output_efficiency_inst='+str(round(math.sqrt(RTE[i1-1]),6))

     wacc_instance = ' --wacc_instance=0.055'                    
     equity_perc_inst = ' --perc_equity_instance=' + str(round(equity_percentage,4))
     ror_inst = ' --ror_instance=0.0489'
     roe_inst = ' --roe_instance=0.104'
     debt_interest_inst = ' --debt_interest_instance=0.0489'
     cftr_inst = ' --cftr_instance=0.27'
     inflation_inst = ' --inflation_inst=' + str(round(inflation_rate,3))
     bonus_dep_frac_inst = ' --bonus_deprec_instance=' + str(round(bonus_depreciation,1))
     
     storage_init_inst = ' --storage_init_instance=0.5'
     storage_final_inst = ' --storage_final_instance=0.5'
     max_storage_dur_inst= ' --max_stor_disch_inst=10000'
     
     storage_cap = ' --storage_cap_instance='+str(h2_storage_duration)#'1000'#+str(stor_dur[i1-1])
     storage_opt = ' --opt_storage_cap ='+str(optimize_storage_duration)
     out_dir = ' --outdir='+dirout
     in_dir = ' --indir='+dir1
     #out_dir = ' --outdir=C:\\Users\\ereznic2\\Documents\\Projects\\SCS_CRADA\\RODeO\\Projects\\SCS\\Output_GSA_test'
     #in_dir = ' --indir=C:\\Users\\ereznic2\\Documents\\Projects\\SCS_CRADA\\RODeO\\Projects\\SCS\\Data_files\\TXT_files'
     product_price_inst = ' --Product_price_instance='+str(lcoh_guessvalue)
     device_ren_inst = ' --devices_ren_instance=1'
     input_cap_inst = ' --input_cap_instance='+str(electrolyzer_size_mw)#1'
     allow_import_inst = ' --allow_import_instance='+str(grid_imports)
     input_LSL_inst = ' --input_LSL_instance=0'
     
     # If grid only, set renewable costs and capacity to zero
     if grid_connection_scenario == 'grid-only':
         ren_cap = ' --Renewable_MW_instance=0'
         ren_capcost = ' --renew_cap_cost_inst=0'
         ren_fom = ' --renew_FOM_cost_inst=0'
     else:
         ren_cap = ' --Renewable_MW_instance='+str(system_rating_mw)#'1'
         ren_capcost = ' --renew_cap_cost_inst='+str(round(hybrid_installed_cost_perMW))#'1230000'
         ren_fom = ' --renew_FOM_cost_inst='+str(1000*round(wind_om_cost_kw))
         
     ren_vom = ' --renew_VOM_cost_inst=0'
     
     input_capcost= ' --input_cap_cost_inst='+str(round(electrolyzer_capex_USD_per_MW))#'1542000'
     prodstor_capcost = ' --ProdStor_cap_cost_inst='+str(round(h2_storage_cost_USDperkg))#'26'
     input_fom = ' --input_FOM_cost_inst='+str(round(electrolyzer_fixed_opex_USD_per_MW_year))#'34926.3'
     input_vom = ' --input_VOM_cost_inst='+str(round(total_variable_OM,2))
     
     water_charge_inst = ' --water_charge_inst='+str(round(water_cost,5))
     desal_cap_cost_inst = ' --desal_cap_cost_inst='+str(round(desal_capex_per_mw))
     desal_FOM_inst = ' --desal_FOM_inst='+str(round(desal_opex_per_mw))
     desal_sys_size_inst = ' --desal_sys_size_inst='+str(round(desal_sys_size_mw,4))
     
     
     ren_itc = ' --itc_ren_inst='+str(round(policy_scenario['Wind ITC'],5))
     stor_itc = ' --itc_stor_inst='+str(round(policy_scenario['Storage ITC'],5))
     
     if atb_year == 2035:
         ren_itc = ' --itc_ren_inst=0'
         stor_itc = ' --itc_stor_inst=0'
     
     # Create batch file
     batch_string = txt1+scenario_inst+demand_prof+ren_prof+load_prof+energy_price+efficiency+storage_cap+storage_opt+ren_cap+out_dir+in_dir\
                  + product_price_inst+device_ren_inst+input_cap_inst+allow_import_inst+input_LSL_inst+ren_capcost+input_capcost+prodstor_capcost\
                  +ren_fom+input_fom+ren_vom+input_vom+water_charge_inst+desal_sys_size_inst+desal_cap_cost_inst+desal_FOM_inst\
                  + wacc_instance+equity_perc_inst+ror_inst+roe_inst+debt_interest_inst+cftr_inst+inflation_inst+bonus_dep_frac_inst\
                  + storage_init_inst+storage_final_inst  +max_storage_dur_inst + ren_itc + stor_itc                           
     
     #   # # For troubleshooting only
     # with open(os.path.join(dir0, 'Output_batch.bat'), 'w') as OPATH:
     #     OPATH.writelines([batch_string,'\n','pause']) # Remove '\n' and 'pause' if not trouble shooting   
     # os.startfile(r'..\\RODeO\\Output_batch.bat')  
       
     #temp = subprocess.run(batch_string,capture_output = True)
     #print(temp)  
     
     #--------------------------- Post processing ---------------------------------
     
     # Get RODeO results summary (high level outputs such as LCOH, capacity factor, cost breakdown, etc.)
     RODeO_results_summary = pd.read_csv(dirout+'Storage_dispatch_summary_'+scenario_name + '.csv',header = 1,sep=',')
     RODeO_results_summary = RODeO_results_summary.rename(columns = {'Elapsed Time (minutes):':'Parameter',RODeO_results_summary.columns[1]:'Value'}).set_index('Parameter')
     # Put results into a dictionary
     RODeO_results_summary_T = RODeO_results_summary.T
     RODeO_results_summary_dict = RODeO_results_summary_T.iloc[0].to_dict()
    
     # Examples for reading out RODeO summary results of interest
     lcoh = RODeO_results_summary_dict['Product NPV cost (US$/kg)']
     electrolyzer_capacity_factor = RODeO_results_summary_dict['input capacity factor']
     electrolyzer_renewable_curtailment_MWh = RODeO_results_summary_dict['Curtailment (MWh)']
     electyrolyzer_renewable_curtailment_percent = 100*RODeO_results_summary_dict['Curtailment (MWh)']/(RODeO_results_summary_dict['Renewable Electricity Input (MWh)']+0.00000001)
     storage_duration_hr = RODeO_results_summary_dict['storage capacity (MWh)']/RODeO_results_summary_dict['input efficiency (%)']/system_rating_mw
     storage_capacity_kg = RODeO_results_summary_dict['storage capacity (MWh)']/electrolysis_desal_total_energy_consumption*1000
    
     # Get RODeO operational results (e.g., electrolyzer and storage hourly operation)
     hydrogen_hourly_inputs_RODeO = pd.read_csv(dirout+'Storage_dispatch_inputs_'+scenario_name + '.csv',index_col = None,header = 29)
     hydrogen_hourly_results_RODeO = pd.read_csv(dirout+'Storage_dispatch_results_'+scenario_name + '.csv',index_col = None,header = 26)
     hydrogen_hourly_results_RODeO['Storage Level (%)'] = 100*hydrogen_hourly_results_RODeO['Storage Level (MW-h)']/(RODeO_results_summary_dict['storage capacity (MWh)'])
     hydrogen_hourly_results_RODeO['Electrolyzer hydrogen production [kg/hr]'] = hydrogen_hourly_results_RODeO['Input Power (MW)']*1000/54.55
     hydrogen_hourly_results_RODeO['Water consumption [kg/hr]'] = hydrogen_hourly_results_RODeO['Electrolyzer hydrogen production [kg/hr]']*10 #15.5 might be a better number for centralized electrolysis
    
     hydrogen_annual_production = sum(hydrogen_hourly_results_RODeO['Product Sold (units of product)'])
     water_consumption_hourly_array = hydrogen_hourly_results_RODeO['Water consumption [kg/hr]'].to_numpy()
     
     return(scenario_name,lcoh,electrolyzer_capacity_factor,storage_duration_hr,storage_capacity_kg,hydrogen_annual_production,water_consumption_hourly_array,RODeO_results_summary_dict,hydrogen_hourly_results_RODeO,electrical_generation_timeseries,electrolyzer_system_capex_kw,h2_storage_cost_USDperkg)


import os

from pytest import approx

from greenheart.simulation.greenheart_simulation import (
    run_simulation,
    GreenHeartSimulationConfig,
)

dirname = os.path.dirname(__file__)
path = os.path.join(dirname, "input/")


def test_onshore_steel_mn_2030_no_policy(subtests):
    # load inputs as needed
    turbine_model = "lbw_6MW"
    filename_turbine_config = os.path.join(path, f"turbines/{turbine_model}.yaml")
    filename_floris_config = os.path.join(path, "floris/floris_input_lbw_6MW.yaml")
    filename_hopp_config = os.path.join(path, "plant/hopp_config_mn_2030.yaml")
    filename_greenheart_config = os.path.join(
        path, "plant/greenheart_config_onshore_mn_2030.yaml"
    )

    config = GreenHeartSimulationConfig(
        filename_hopp_config,
        filename_greenheart_config,
        filename_turbine_config,
        filename_floris_config,
        verbose=True,
        show_plots=False,
        save_plots=True,
        use_profast=True,
        post_processing=True,
        incentive_option=1,
        plant_design_scenario=9,
        output_level=8,
    )

    output = run_simulation(config)

    with subtests.test("aep"):
        assert output.hopp_results["hybrid_plant"].annual_energies["wind"] == approx(
            4007165638.0913243, rel=0.1
        )

    with subtests.test("wind cf"):
        assert output.hopp_results["hybrid_plant"].capacity_factors["wind"] == approx(
            35.3, rel=0.1
        )

    with subtests.test("h2 capacity factor"):
        assert output.electrolyzer_physics_results["H2_Results"][
            "Life: Capacity Factor"
        ] == approx(0.3817629024505595, rel=0.1)

    with subtests.test("electrolyzer installed capex"):
        assert output.capex_breakdown["electrolyzer"] / 1160000 == approx(
            553.4592000000001, rel=0.1
        )

    with subtests.test("lcoh"):
        assert output.lcoh == approx(4.2986685034417045, rel=0.1)

    with subtests.test("h2 production"):
        assert output.electrolyzer_physics_results["H2_Results"][
            "Life: Annual H2 production [kg/year]"
        ] == approx(71527565.78081538, rel=0.1)

    with subtests.test("capex"):
        assert sum(output.capex_breakdown.values()) == approx(2603580420.446765, 0.1)

    with subtests.test("wind capex"):
        assert output.capex_breakdown["wind"] == approx(1706186592.0000002, rel=0.1)

    # Not included in test suite. Difference between GS and GreenHEART
    # with subtests.test('compressor capex'):
    #     assert output.capex_breakdown['h2_transport_compressor'] == approx(18749265.76357594, rel=0.1)

    with subtests.test("electrolyzer capex"):
        assert output.capex_breakdown["electrolyzer"] == approx(
            642012672.0000001, rel=0.1
        )
    # Not included in test suite. Difference between GS and GreenHEART
    # with subtests.test("h2 storage capex"):
    #     assert output.capex_breakdown["h2_storage"] == approx(
    #         234630860.95910057, rel=0.1
    #     )

    # Not included in test suite. Difference between GS and GreenHEART
    # with subtests.test('desal capex'):
    #     assert output.capex_breakdown['desal'] == approx(294490.93738406134, rel=0.1)

    with subtests.test("lcos"):
        assert output.steel_finance.sol.get("price") == approx(
            961.2866791076059, rel=0.1
        )


def test_onshore_ammonia_tx_2030_no_policy(subtests):
    # load inputs as needed
    turbine_model = "lbw_6MW"
    filename_turbine_config = os.path.join(path, f"turbines/{turbine_model}.yaml")
    filename_floris_config = os.path.join(path, "floris/floris_input_lbw_6MW.yaml")
    filename_hopp_config = os.path.join(path, "plant/hopp_config_tx_2030.yaml")
    filename_greenheart_config = os.path.join(
        path, "plant/greenheart_config_onshore_tx_2030.yaml"
    )

    config = GreenHeartSimulationConfig(
        filename_hopp_config,
        filename_greenheart_config,
        filename_turbine_config,
        filename_floris_config,
        verbose=True,
        show_plots=False,
        save_plots=True,
        use_profast=True,
        post_processing=True,
        incentive_option=1,
        plant_design_scenario=9,
        output_level=8,
    )

    output = run_simulation(config)

    with subtests.test("lcoh"):
        assert output.lcoh == approx(4.023963541118758, rel=0.1)

    with subtests.test("lcoa"):
        assert output.ammonia_finance.sol.get("price") == approx(
            0.8839797787889466, rel=0.1
        )

    with subtests.test("aep"):
        assert output.hopp_results["hybrid_plant"].annual_energies["wind"] == approx(
            3383439.9656016766 * 1e3, rel=0.1
        )

    with subtests.test("wind cf"):
        assert output.hopp_results["hybrid_plant"].capacity_factors["wind"] == approx(
            36.78, rel=0.1
        )

    with subtests.test("h2 capacity factor"):
        assert output.electrolyzer_physics_results["H2_Results"][
            "Life: Capacity Factor"
        ] == approx(0.38749, rel=0.1)

    with subtests.test("electrolyzer installed capex"):
        assert output.capex_breakdown["electrolyzer"] / 960000 == approx(
            553.4592000000001, rel=0.1
        )

    with subtests.test("h2 production"):
        assert output.electrolyzer_physics_results["H2_Results"][
            "Life: Annual H2 production [kg/year]"
        ] == approx(60400414.51854633, rel=0.1)

    with subtests.test("capex"):
        assert sum(output.capex_breakdown.values()) == approx(
            1922324028.6785245, rel=0.1
        )

    with subtests.test("wind capex"):
        assert output.capex_breakdown["wind"] == approx(1196393976.0, rel=0.1)

    # Not included in test suite. Difference between GS and GreenHEART
    # with subtests.test('compressor capex'):
    #     assert output.capex_breakdown['h2_transport_compressor'] == approx(14082368.770328939, rel=0.1)

    with subtests.test("electrolyzer capex"):
        assert output.capex_breakdown["electrolyzer"] == approx(
            531320832.0000001, rel=0.1
        )
    # Not included in test suite. Difference between GS and GreenHEART
    # with subtests.test("h2 storage capex"):
    #     assert output.capex_breakdown["h2_storage"] == approx(
    #         178870827.30895007, rel=0.1
    #     )
    # Not included in test suite. Difference between GS and GreenHEART
    # with subtests.test("desal capex"):
    #     assert output.capex_breakdown["desal"] == approx(1656024.5992454647, rel=0.1)


def test_onshore_ammonia_tx_2030_base_policy(subtests):
    # load inputs as needed
    # NOTE: GreenSteel used Wind PTC, H2 PTC and Storage ITC for base policy and incentive_option 2 does not include the Storage ITC
    turbine_model = "lbw_6MW"
    filename_turbine_config = os.path.join(path, f"turbines/{turbine_model}.yaml")
    filename_floris_config = os.path.join(path, "floris/floris_input_lbw_6MW.yaml")
    filename_hopp_config = os.path.join(path, "plant/hopp_config_tx_2030.yaml")
    filename_greenheart_config = os.path.join(
        path, "plant/greenheart_config_onshore_tx_2030.yaml"
    )

    config = GreenHeartSimulationConfig(
        filename_hopp_config,
        filename_greenheart_config,
        filename_turbine_config,
        filename_floris_config,
        verbose=True,
        show_plots=False,
        save_plots=True,
        use_profast=True,
        post_processing=True,
        incentive_option=2,
        plant_design_scenario=9,
        output_level=8,
    )

    output = run_simulation(config)

    with subtests.test("lcoh"):
        assert output.lcoh == approx(
            3.2231088737405846,
            rel=0.11
        )

    with subtests.test("lcoa"):
        assert output.ammonia_finance.sol.get("price") == approx(
            0.7242460501735473,
             rel=0.1
        )

    with subtests.test("aep"):
        assert output.hopp_results["hybrid_plant"].annual_energies["wind"] == approx(
            3383439.9656016766 * 1e3, 
            rel=0.1
        )

    with subtests.test("wind cf"):
        assert output.hopp_results["hybrid_plant"].capacity_factors["wind"] == approx(
            36.78, rel=0.1
        )

    with subtests.test("h2 capacity factor"):
        assert output.electrolyzer_physics_results["H2_Results"][
            "Life: Capacity Factor"
        ] == approx(0.4023306656204429, rel=0.1)

    with subtests.test("electrolyzer installed capex"):
        assert output.capex_breakdown["electrolyzer"] / 960000 == approx(
            553.4592000000001, rel=0.1
        )

    with subtests.test("h2 production"):
        assert output.electrolyzer_physics_results["H2_Results"][
            "Life: Annual H2 production [kg/year]"
        ] == approx(60400414.51854633, rel=0.1)

    with subtests.test("capex"):
        assert sum(output.capex_breakdown.values()) == approx(1922324028.6785245, 
                                                              0.1
                                                              )


def test_onshore_ammonia_tx_2030_max_policy(subtests):
    # load inputs as needed
    # NOTE: GreenSteel used Wind PTC, H2 PTC and Storage ITC for max policy and incentive_option 3 does not include the Storage ITC
    turbine_model = "lbw_6MW"
    filename_turbine_config = os.path.join(path, f"turbines/{turbine_model}.yaml")
    filename_floris_config = os.path.join(path, "floris/floris_input_lbw_6MW.yaml")
    filename_hopp_config = os.path.join(path, "plant/hopp_config_tx_2030.yaml")
    filename_greenheart_config = os.path.join(
        path, "plant/greenheart_config_onshore_tx_2030.yaml"
    )

    config = GreenHeartSimulationConfig(
        filename_hopp_config,
        filename_greenheart_config,
        filename_turbine_config,
        filename_floris_config,
        verbose=True,
        show_plots=False,
        save_plots=True,
        use_profast=True,
        post_processing=True,
        incentive_option=7,
        plant_design_scenario=9,
        output_level=8,
    )

    output = run_simulation(config)

    with subtests.test("lcoh"):
        assert output.lcoh == approx(-0.3719231321573331, rel=10)

    with subtests.test("capex"):
        assert sum(output.capex_breakdown.values()) == approx(1922324028.6785245, 0.1)

    with subtests.test("lcoa"):
        assert output.ammonia_finance.sol.get("price") == approx(
            0.007202259061804786, rel=100
        )

    with subtests.test("aep"):
        assert output.hopp_results["hybrid_plant"].annual_energies["wind"] == approx(
            3383439.9656016766 * 1e3, rel=0.1
        )

    with subtests.test("wind cf"):
        assert output.hopp_results["hybrid_plant"].capacity_factors["wind"] == approx(
            36.78, rel=0.1
        )

    with subtests.test("h2 capacity factor"):
        assert output.electrolyzer_physics_results["H2_Results"][
            "Life: Capacity Factor"
        ] == approx(0.3874913813926033, rel=0.1)

    with subtests.test("electrolyzer installed capex"):
        assert output.capex_breakdown["electrolyzer"] / 960000 == approx(
            553.4592000000001, rel=0.1
        )

    with subtests.test("h2 production"):
        assert output.electrolyzer_physics_results["H2_Results"][
            "Life: Annual H2 production [kg/year]"
        ] == approx(60400414.51854633, rel=0.1)

    with subtests.test("capex"):
        assert sum(output.capex_breakdown.values()) == approx(1922324028.6785245, 0.1)


def test_onshore_ammonia_tx_2025_no_policy(subtests):
    # load inputs as needed
    turbine_model = "lbw_6MW"
    filename_turbine_config = os.path.join(path, f"turbines/{turbine_model}.yaml")
    filename_floris_config = os.path.join(path, "floris/floris_input_lbw_6MW.yaml")
    filename_hopp_config = os.path.join(path, "plant/hopp_config_tx_2025.yaml")
    filename_greenheart_config = os.path.join(
        path, "plant/greenheart_config_onshore_tx_2025.yaml"
    )

    config = GreenHeartSimulationConfig(
        filename_hopp_config,
        filename_greenheart_config,
        filename_turbine_config,
        filename_floris_config,
        verbose=True,
        show_plots=False,
        save_plots=True,
        use_profast=True,
        post_processing=True,
        incentive_option=1,
        plant_design_scenario=9,
        output_level=8,
    )

    output = run_simulation(config)

    with subtests.test("lcoh"):
        assert output.lcoh == approx(5.215316325627814, rel=0.1)

    with subtests.test("lcoa"):
        assert output.ammonia_finance.sol.get("price") == approx(
            1.1211844274252674, rel=0.1
        )

    with subtests.test("aep"):
        assert output.hopp_results["hybrid_plant"].annual_energies["wind"] == approx(
            3383439.9656016766 * 1e3, rel=0.1
        )

    with subtests.test("wind cf"):
        assert output.hopp_results["hybrid_plant"].capacity_factors["wind"] == approx(
            36.78, rel=0.1
        )

    with subtests.test("h2 capacity factor"):
        assert output.electrolyzer_physics_results["H2_Results"][
            "Life: Capacity Factor"
        ] == approx(0.3874913813926033, rel=0.1)

    with subtests.test("electrolyzer installed capex"):
        assert output.capex_breakdown["electrolyzer"] / 960000 == approx(
            923.3862399999999, rel=0.1
        )

    with subtests.test("h2 production"):
        assert output.electrolyzer_physics_results["H2_Results"][
            "Life: Annual H2 production [kg/year]"
        ] == approx(60400414.51854633, rel=0.1)

    with subtests.test("capex"):
        assert sum(output.capex_breakdown.values()) == approx(2400580143.0785246, 0.1)

    with subtests.test("h2 production"):
        assert output.electrolyzer_physics_results["H2_Results"][
            "Life: Annual H2 production [kg/year]"
        ] == approx(60400414.51854633, rel=0.1)

    with subtests.test("wind capex"):
        assert output.capex_breakdown["wind"] == approx(1319520132.0000002, rel=0.1)

    with subtests.test("electrolyzer capex"):
        assert output.capex_breakdown["electrolyzer"] == approx(
            886450790.3999999, rel=0.1
        )

    # Not included in test suite. Difference between GS and GreenHEART
    # with subtests.test("h2 storage capex"):
    #     assert output.capex_breakdown["h2_storage"] == approx(
    #         178870827.30895007, rel=0.1
    #     )

    # Not included in test suite. Difference between GS and GreenHEART
    # with subtests.test('desal capex'):
    #     assert output.capex_breakdown['desal'] == approx(294490.93738406134, rel=0.1)

    # Not included in test suite. Difference between GS and GreenHEART
    # with subtests.test('compressor capex'):
    #     assert output.capex_breakdown['h2_transport_compressor'] == approx(18749265.76357594, rel=0.1)

def test_onshore_ammonia_tx_2035_no_policy(subtests):
    # load inputs as needed
    turbine_model = "lbw_6MW"
    filename_turbine_config = os.path.join(path, f"turbines/{turbine_model}.yaml")
    filename_floris_config = os.path.join(path, "floris/floris_input_lbw_6MW.yaml")
    filename_hopp_config = os.path.join(path, "plant/hopp_config_tx_2035.yaml")
    filename_greenheart_config = os.path.join(
        path, "plant/greenheart_config_onshore_tx_2035.yaml"
    )

    config = GreenHeartSimulationConfig(
        filename_hopp_config,
        filename_greenheart_config,
        filename_turbine_config,
        filename_floris_config,
        verbose=True,
        show_plots=False,
        save_plots=True,
        use_profast=True,
        post_processing=True,
        incentive_option=1,
        plant_design_scenario=9,
        output_level=8,
    )

    output = run_simulation(config)

    with subtests.test("lcoh"):
        assert output.lcoh == approx(3.68491615716891, rel=0.1)

    with subtests.test("lcoa"):
        assert output.ammonia_finance.sol.get("price") == approx(
            0.8161748484435717, rel=0.1
        )

    with subtests.test("aep"):
        assert output.hopp_results["hybrid_plant"].annual_energies["wind"] == approx(
            3383439.9656016766 * 1e3, rel=0.1
        )

    with subtests.test("wind cf"):
        assert output.hopp_results["hybrid_plant"].capacity_factors["wind"] == approx(
            36.78, rel=0.1
        )

    with subtests.test("h2 capacity factor"):
        assert output.electrolyzer_physics_results["H2_Results"][
            "Life: Capacity Factor"
        ] == approx(0.3874913813926033, rel=0.1)

    with subtests.test("electrolyzer installed capex"):
        assert output.capex_breakdown["electrolyzer"] / 960000 == approx(
            457.55808, rel=0.1
        )

    with subtests.test("h2 production"):
        assert output.electrolyzer_physics_results["H2_Results"][
            "Life: Annual H2 production [kg/year]"
        ] == approx(60400414.51854633, rel=0.1)

    with subtests.test("capex"):
        assert sum(output.capex_breakdown.values()) == approx(1771419905.4785244, 0.1)

    with subtests.test("wind capex"):
        assert output.capex_breakdown["wind"] == approx(1137554928.0, rel=0.1)

    with subtests.test("electrolyzer capex"):
        assert output.capex_breakdown["electrolyzer"] == approx(439255756.8, rel=0.1)

    # Not included in test suite. Difference between GS and GreenHEART
    # with subtests.test("h2 storage capex"):
    #     assert output.capex_breakdown["h2_storage"] == approx(
    #         178870827.30895007, rel=0.1
    #     )

    # Not included in test suite. Difference between GS and GreenHEART
    # with subtests.test('desal capex'):
    #     assert output.capex_breakdown['desal'] == approx(294490.93738406134, rel=0.1)

    # Not included in test suite. Difference between GS and GreenHEART
    # with subtests.test('compressor capex'):
    #     assert output.capex_breakdown['h2_transport_compressor'] == approx(18749265.76357594, rel=0.1)



from scipy.optimize import fsolve
import numpy as np
import pandas as pd

def get_diameter_of_pipe(L:float,m_dot:float,p_inlet:float,p_outlet:float):
    '''
    Overview:
    ---------
        This function returns the diameter of a pipe for a given length,flow rate, and pressure boundaries

    Parameters:
    -----------
        L : float - Length of pipeline [km]
        m_dot : float = Mass flow rate [kg/s]
        p_inlet : float = Pressure at inlet of pipe [bar]
        p_outlet : float = Pressure at outlet of pipe [bar]

    Returns:
    --------
        diameter_in : float - Diameter of pipe [inches]

    '''

    #   Conversions
    m2in = 39.3701

    #   Various inputs
    f = 0.01                # Friction factor
    Z = 0.9                 # Compressibility
    R = 8.314/(2.016/1000)  # J/kg-K For hydrogen
    T = 15+273              # Temperature [K]
    nodes = 50              # Number of nodes to discretize with
    params = (m_dot,p_inlet,p_outlet,L,Z*R*T,f,nodes)

    #   Generate guess values
    d_init = 0.01           # Diameter [m]
    p_init = np.linspace(p_inlet,p_outlet,nodes-1) #[bar]
    x_init = np.concatenate(([d_init],p_init))

    #   Solve for the diameter
    diameter_m = fsolve(momentum_bal,x_init,args = params)[0]
    diameter_in = diameter_m * m2in

    return diameter_in


def momentum_bal(x,*params):
    '''
    Overview:
    ---------
        Residual equation for momentum balance

    Parameters:
    -----------
        x : list[float] - Guess values. Index 0 is the diameter. The remaining are the pressures
        params : tuple(float) - Parameters in the order m_dot,p_inlet,p_outlet,L_km,ZRT,f,nodes

    Returns:
    --------
        residual : list[float] - Residual values of momentum equation
    '''
    
    #   Useful conversions
    bar2Pa = 100000     #   Convert bar to Pa 
    km2m = 1000         #   Convert km to m

    # Unload params into local varibles
    m_dot,p_inlet,p_outlet,L,ZRT,f,nodes = params 

    #   Extract guess values
    D = x[0]                                                    #   Diameter of pipe [m]
    p_bar = x[1:]                                               #   Pressure inside the pipe [bar]

    #   Geometric parameters
    A = np.pi/4*D**2                                            #   Calculate area of pipe [m2]
    #L = L_km * km2m                                             #   Length of pipe [m]
    dz = L/nodes                                                #   Uniform spacing of nodes [m]

    #   Pressure, density, and velocity for each node
    p = np.concatenate(([p_inlet],p_bar,[p_outlet]))*bar2Pa     #   Pressure array with boundary conditions [Pa]
    rho = p/ZRT                                                 #   Density of gas [kg/m3]
    rho_avg = np.mean([rho[0:-1],rho[1:]],axis=0)               #   Average density in each node [kg/m3]
    v = m_dot/rho/A                                             #   Velocity of gas [m/s]
    v_avg = np.mean([v[0:-1],v[1:]],axis=0)                     #   Average density in each node [m/s]

    #   Momentum balance terms
    dpdz = (p[1:]-p[0:-1])/dz                                   #   Pressure drop [Pa/m]
    tau = 0.5*f*rho_avg*v_avg**2                                #   Viscous drag [Pa]
    pududz = rho_avg*v_avg * (v[1:]-v[0:-1])/dz                 #   Velocity change effect [Pa/m]

    #   Momentum balance
    residual = -dpdz - tau/D - pududz 

    return residual

def get_dn_schedule_of_pipe(pipe_info_dir,grade,design_option,location_class,joint_factor,pipe_inner_diameter,design_pressure_asme):
    # Read in yield strengths for API 5L steels
    yield_strengths = pd.read_csv(pipe_info_dir+'steel_mechanical_props.csv',index_col = None,header = 0)

    # Isolate the yield strength of the grade selecdted
    yield_strength = yield_strengths.loc[yield_strengths['Grade']==grade,'SMYS [Mpa]'].to_list()[0]

    # Assign design factor based on design option
    if design_option == 'a' or design_option == 'A':
        design_factor_list = [0.5,0.5,0.5,0.4]
    elif design_option == 'b' or design_option == 'B':
        design_factor_list = [0.72,0.60,0.50,0.40]
    else:
        design_factor = [0.5,0.5,0.5,0.4]
        print('Error. Invalid design option. Option set to a as default. Please enter a, A, b, or B for design option')

    # Determine design factor based on location class
    design_factor = design_factor_list[location_class - 1]

    # Design pressure array for ASME B31.12 Table for material performance factor
    dp_array = np.array([6.8948,13.7895,15.685,16.5474,17.9264,19.3053,20.6843])
    
    # Read in possible pipe schedules for different nominal diameters in metric units
    pipe_schedules = pd.read_csv(pipe_info_dir+'pipe_dimensions_metric.csv',index_col = None,header = 0)
    
    # Isolate the nominal diameter options
    dn_options = pipe_schedules['DN'].to_list()
    
    for dn in dn_options:
    
        #dn = dn_options[9]
            
        # Clean up table leaving what is actually needed
        pipe_schedules_DN = pipe_schedules.loc[(pipe_schedules['DN'] == dn)].drop(labels = ['DN','Outer diameter [mm]'], axis = 1).transpose().reset_index()
        pipe_schedules_DN = pipe_schedules_DN.rename(columns = {pipe_schedules_DN.columns[0]:'Schedule',pipe_schedules_DN.columns[1]:'Wall thickness [mm]'}).dropna()
        pipe_schedules_DN = pipe_schedules_DN.reset_index().drop(labels = ['index'],axis=1)
        
        pipe_outer_diameter = pipe_schedules.loc[(pipe_schedules['DN']==dn),'Outer diameter [mm]'].values[0]
        
        if pipe_outer_diameter <= pipe_inner_diameter:
            continue
        else:
        
            if design_option == 'b' or design_option == 'B':
                mat_perf_factor = 1
            else: 
                if yield_strength <= 358.528:
                    h_f_array = np.array([1,1,0.954,0.91,0.88,0.84,0.78])
                elif yield_strength > 358.528 and yield_strength <=413.686:
                    h_f_array = np.array([0.874,0.874,0.834,0.796,0.77,0.734,0.682])
                elif yield_strength > 413.686 and yield_strength <= 482.633:
                    h_f_array = np.array([0.776,0.776,0.742,0.706,0.684,0.652,0.606])
                elif yield_strength > 482.633 and yield_strength<= 551.581:
                    h_f_array = np.array([0.694,0.694,0.662,0.632,0.61,0.584,0.542]) 
                mat_perf_factor = np.interp(design_pressure_asme,dp_array,h_f_array)
            #thickness.append(design_pressure_asme*diam_outer[0]/(2*yield_strengths.loc[k,'SMYS [Mpa]']*design_factor*joint_factor*mat_perf_factor))
            thickness = design_pressure_asme*pipe_outer_diameter/(2*yield_strength*design_factor*joint_factor*mat_perf_factor)
            if max(pipe_schedules_DN['Wall thickness [mm]']) >= thickness:
                schedule_minviable=pipe_schedules_DN.loc[pipe_schedules_DN["Wall thickness [mm]"] == min(pipe_schedules_DN.loc[pipe_schedules_DN['Wall thickness [mm]'] >= thickness,'Wall thickness [mm]']),'Schedule'].to_list()[0]
                thickness_of_schedule = pipe_schedules_DN.loc[pipe_schedules_DN['Schedule']==schedule_minviable,'Wall thickness [mm]'].to_list()[0]
            else:
                continue
                
            inner_diameter_min_viable_schedule = pipe_outer_diameter-2*thickness_of_schedule
            
            if inner_diameter_min_viable_schedule > pipe_inner_diameter:
                break
    
    
            
    return(dn,pipe_outer_diameter,schedule_minviable,thickness_of_schedule)


# if __name__ == "__main__":
#     L = 8                   # Length [km]
#     m_dot = 0.12            # Mass flow rate [kg/s]
#     p_inlet = 30            # Inlet pressure [bar]
#     p_outlet = 10           # Outlet pressure [bar]
#     min_diam_in = get_diameter_of_pipe(L,m_dot,p_inlet,p_outlet)
#     print(f'The minimum diameter is {round(min_diam_in,3)} inches')


"""
Author: Christopher Bay
Date: 01/24/2023
Institution: National Renewable Energy Laboratory
Description: This file implements electrolzyer CapEx and OpEx models from [1]. The exact extent of what is 
             included in the costs is unclear in [1]. Source [2] (cited by [1]) states that 
             "equipment costs include the electrolyser system, the filling centre or compressor skids and storage 
             systems". 
Sources:
    - [1] Singlitico, Alessandro, Jacob Østergaard, and Spyros Chatzivasileiadis. "Onshore, offshore or
        in-turbine electrolysis? Techno-economic overview of alternative integration designs for green hydrogen
        production into Offshore Wind Power Hubs." Renewable and Sustainable Energy Transition 1 (2021): 100005.
    - [2] [E. Tractebel , H. Engie , Study on early business cases for h2 in energy storage and more 
        broadly power to h2 applications, EU Comm, 2017, p. 228 .] 
        https://hsweb.hs.uni-hamburg.de/projects/star-formation/hydrogen/P2H_Full_Study_FCHJU.pdf
"""

import numpy as np
import matplotlib.pyplot as plt


class PEMCostsSingliticoModel():

    def __init__(
        self,
        elec_location: int,
    ):
        """
        Initialize object for PEM costs based on [1].

        Args:
            elec_location (int): Parameter for indicating the electrolyzer location;
                0 is for onshore, 1 is for offshore or in-turbine.
        """
        # Values for CapEX & OpEx taken from [1], Table B.2, PEMEL.
        # Installation costs include land, contingency, contractors, legal fees, construction,
        # engineering, yard improvements, buildings, electrics, piping, instrumentation,
        # and installation and grid connection.
        self.IF = 0.33 # instatllation fraction [% RC_elec]
        self.RP_elec = 10 # reference power [MW]

        # Values for OpEx taken from [1], Table B.3, PEMEL.
        self.RP_SR = 5 # reference power [MW]
        self.RU_SR = 0.41 # reference cost share [%], for a reference power, RP_SR, of 5MW
        self.P_stack_max_bar = 2 # average max size [MW]
        self.SF_SR_0 = 0.11 # average scale factor


        self.OS = elec_location # 1 for offshore or in-turbine electrolyzer location, 0 for onshore; from [1],
                                # Table B.1 notes for CapEx_el
        
        # NOTE: This is used in the stack replacement cost code that is currently commented out; more work needs
        # to be done to make sure this is set and used correctly.
        # self.P_elec_bar = 1 * 10**3 # scaled maximum [MW] from [1], Table B.1 notes for OpEx_elec_eq

        # NOTE: This is used in the stack replacement cost code that is currently commented out.
        # self.OH_max = 85000 # Lifetime maximum operating hours [h], taken from [1], Table 1, PEMEL

    def run(
        self,
        P_elec: float,
        RC_elec: float,
    ) -> tuple:
        """
        Computes the CapEx and OpEx costs for a single electrolyzer.

        Args:
            P_elec (float): Nominal capacity of the electrolyzer [GW].
            RC_elec (float): Reference cost of the electrolyzer [MUSD/GW] for a 10 MW electrolyzer plant installed.

        Returns:
            tuple: CapEx and OpEx costs for a single electrolyzer.
        """
        capex = self.calc_capex(P_elec, RC_elec)
        
        opex = self.calc_opex(P_elec, capex)

        return capex, opex

    def calc_capex(
        self,
        P_elec: float,
        RC_elec: float,
    ) -> float:
        """
        CapEx for a single electrolyzer, given the electrolyzer capacity and reference cost.
        Equation from [1], Table B.1, CapEx_EL. For in-turbine electrolyzers,
        it is assumed that the maximum electrolyzer size is equal to the turbine rated capacity.

        NOTE: If the single electrolyzer capacity exceeds 100MW, the CapEx becomes fixed at the cost of a
        100MW system, due to decreasing economies of scale (based on assumption from [1]).
        As such, if you use the output to calculate a cost per unit of electrolyzer, you will need to divide
        the cost by 100MW and not the user-specified size of the electrolyzer for sizes above 100 MW.

        Args:
            P_elec (float): Nominal capacity of the electrolyzer [GW].
            RC_elec (float): Reference cost of the electrolyzer [MUSD/GW].

        Returns:
            float: CapEx for electrolyzer [MUSD].
        """
        # Choose the scale factor based on electrolyzer size, [1], Table B.2.
        if P_elec < 10 / 10**3:
            self.SF_elec = -0.21 # scale factor, -0.21 for <10MW, -0.14 for >10MW
        else:
            self.SF_elec = -0.14 # scale factor, -0.21 for <10MW, -0.14 for >10MW
        
        # If electrolyzer capacity is >100MW, fix unit cost to 100MW electrolyzer as economies of scale
        # stop at sizes above this, according to assumption in [1].
        if P_elec > 100 / 10**3:
            P_elec_cost_per_unit_calc = 0.1
        else:
            P_elec_cost_per_unit_calc = P_elec

        # Return the cost of a single electrolyzer of the specified capacity in millions of USD (or the supplied currency).
        # MUSD = GW   * MUSD/GW *           -             *      GW   * MW/GW /      MW       **      -
        cost = P_elec_cost_per_unit_calc * RC_elec * (1 + self.IF * self.OS) *  ((P_elec_cost_per_unit_calc * 10**3 / self.RP_elec) ** self.SF_elec)
        cost_per_unit = cost / P_elec_cost_per_unit_calc

        return cost_per_unit * P_elec


    def calc_opex(
        self,
        P_elec: float,
        capex_elec: float,
        RC_elec: float = None,
        OH: float = None,
    ) -> float:
        """
        OpEx for a single electrolyzer, given the electrolyzer capacity and reference cost.
        Equations from [1], Table B.1, OpEx_elec_eq and OpEx_elec_neq.
        The returned OpEx cost include equipment and non-equipment costs, but excludes the stack replacement cost.

        NOTE: If the single electrolyzer capacity exceeds 100MW, the OpEx becomes fixed at the cost of a
        100MW system, due to decreasing economies of scale (based on assumption from [1]).
        As such, if you use the output to calculate a cost per unit of electrolyzer, you will need to divide
        the cost by 100MW and not the user-specified size of the electrolyzer for sizes above 100 MW.

        NOTE: Code for the stack replacement cost is included below, but does not currently match results
        from [1]. DO NOT USE in the current form.

        Args:
            P_elec (float): Nominal capacity of the electrolyzer [GW].
            capex_elec (float): CapEx for electrolyzer [MUSD].
            RC_elec (float, optional): Reference cost of the electrolyzer [MUSD/GW]. Defaults to None. Not currently used.
            OH (float, optional): Operating hours [h]. Defaults to None. Not currently used.

        Returns:
            float: OpEx for electrolyzer [MUSD].
        """
        # If electrolyzer capacity is >100MW, fix unit cost to 100MW electrolyzer as economies of scale
        # stop at sizes above this, according to assumption in [1].
        if P_elec > 100 / 10**3:
            P_elec = 0.1

        # Including material cost for planned and unplanned maintenance, labor cost in central Europe, which
        # all depend on a system scale. Excluding the cost of electricity and the stack replacement,
        # calculated separately. Scaled maximum to P_elec_bar = 1 GW.
        # MUSD*MW         MUSD    *              -                *    -   *    GW   * MW/GW
        opex_elec_eq = capex_elec * (1 - self.IF * (1 + self.OS)) * 0.0344 * (P_elec * 10**3) ** -0.155

        # Covers the other operational expenditure related to the facility level. This includes site
        # management, land rent and taxes, administrative fees (insurance, legal fees...), and site maintenance.
        # MUSD                    MUSD     
        opex_elec_neq = 0.04 * capex_elec * self.IF * (1 + self.OS)

        # NOTE: The stack replacement costs below  don't match the results in [1] supplementary materials.
        # ***DO NOT USE*** stack replacement cost in its current form.

        # Choose the scale factor based on electrolyzer size, [1], Table B.2.
        # if P_elec < 10 / 10**3:
        #     self.SF_elec = -0.21 # scale factor, -0.21 for <10MW, -0.14 for >10MW
        # else:
        #     self.SF_elec = -0.14 # scale factor, -0.21 for <10MW, -0.14 for >10MW

        # Approximation of stack costs and replacement cost depending on the electrolyzer equipment costs.
        # Paid only the year in which the replacement is needed.
        # MUSD/GW    %     * MUSD/GW *       -       *      MW     /      MW       **       -   
        # RC_SR = self.RU_SR * RC_elec * (1 - self.IF) * (self.RP_SR / self.RP_elec) ** self.SF_elec
        # # -                 -          *               MW        /         MW
        # SF_SR = 1 - (1 - self.SF_SR_0) * np.exp(-self.P_elec_bar / self.P_stack_max_bar)
        # # SF_SR = 1 - (1 - self.SF_SR_0) * np.exp(-P_elec * 10**3 / self.P_stack_max_bar)
        # # MUSD           GW   * MUSD/GW *  GW   * MW/GW      MW       **   -   *  h /      h    
        # opex_elec_sr = P_elec * RC_SR * (P_elec * 10**3 / self.RP_SR) ** SF_SR * OH / self.OH_max

        return opex_elec_eq + opex_elec_neq

if __name__ == "__main__":
    # setup up PEM for offshore/in-turbine costs
    pem = PEMCostsSingliticoModel(elec_location=1)

    RC_elec = 600 # reference cost of electrolyzer [MUSD/GW]
    P_elec = 0.000625 # electrolzyer capacity [GW]

    # calculate CapEx and OpEx per unit costs
    capex = pem.calc_capex(P_elec, RC_elec)
    opex = pem.calc_opex(P_elec, capex)

    print('capex [MUSD/GW]: ', capex / P_elec)
    print('opex [MUSD/GW]: ', opex / P_elec)


    # capex = []
    # capex_norm = []
    # opex = []
    # opex_norm = []
    # elec_nom_tot = np.arange(0.5, 12.1, 0.5)
    # # elec_nom = np.linspace(0.000625, 0.015, 24)
    # elec_nom = np.arange(0.5, 12.1, 0.5)
    # # elec_nom = np.arange(0.1, 0.5, 0.1)
    # for nom in elec_nom:
    #     capex.append(pem.calc_capex(nom, RC_elec))
    #     capex_norm.append(pem.calc_capex(nom, RC_elec) / nom)
    #     opex.append(pem.calc_opex(nom, capex[-1]))
    #     opex_norm.append(pem.calc_opex(nom, capex[-1]) / 0.1)

    # # opex_eq = np.array([val[0] for val in opex])
    # # opex_sr = np.array([val[1] for val in opex])
    # # opex_neq = np.array([val[2] for val in opex])

    # # base_capex = [1473.96293,
    # #     1279.788173,
    # #     1173.874669,
    # #     1085.613416,
    # #     1032.656664,
    # #     1050.308914,
    # #     997.3521624,
    # #     979.6999117,
    # #     944.3954104,
    # #     944.3954104,
    # #     926.7431598,
    # #     891.4386584,
    # #     873.7864078,
    # #     873.7864078,
    # #     838.4819064,
    # #     838.4819064,
    # #     838.4819064,
    # #     838.4819064,
    # #     838.4819064,
    # #     838.4819064,
    # #     785.5251545,
    # #     803.1774051,
    # #     803.1774051,
    # #     767.8729038,
    # # ]

    # # fig, ax = plt.subplots(1)
    # # plt.plot(elec_nom_tot, capex)
    # # plt.plot(elec_nom_tot, base_capex)
    # # # plt.plot(elec_nom, np.array(base_capex) * 4.3)
    # # # plt.plot(elec_nom, np.array(capex) * 3)
    # # plt.title('CapEx')
    # # # ax.set_aspect('equal')

    # fig, ax = plt.subplots(1)
    # plt.plot(elec_nom_tot, opex_norm)
    # plt.title('OpEx')
    # # ax.set_aspect('equal')
    # plt.show()

    # # fig, ax = plt.subplots(1)
    # # plt.plot(elec_nom_tot, opex_sr)
    # # plt.title('OpEx SR')
    # # # ax.set_aspect('equal')
    # # plt.show()


    ################################# plot a sweep of sizes for OPEX and CAPEX

    # electrolyzer_capex_kw = 1300 # $/kW
    # time_between_replacement = 65000 # hours
    # electrolyzer_sizes_mw = np.arange(1, 1000)
    # useful_life = 30 # years
    # atb_year = 2025
    # # electrical_generation_timeseries_kw = np.sin(np.arange(0,24*365)*1E-3)*0.5E6 + 0.6E6
    # hydrogen_annual_output = 0

    # # for distributed
    # ndivs = [2, 5, 10, 20]

    # opex = []
    # capex = []
    # opex_distributed = np.zeros((len(ndivs), len(electrolyzer_sizes_mw)))
    # capex_distributed = np.zeros((len(ndivs), len(electrolyzer_sizes_mw)))


    # # centralized
    # pem = PEMCostsSingliticoModel(elec_location=1) # offshore

    # for i, electrolyzer_size_mw in enumerate(electrolyzer_sizes_mw):

    #     electrical_generation_timeseries_kw = electrolyzer_size_mw*1000*np.ones(365*24)


    #     # calculate CapEx and OpEx per unit costs
    #     electrolyzer_total_capital_cost = pem.calc_capex(electrolyzer_size_mw*1E-3, electrolyzer_capex_kw)*1E6
    #     electrolyzer_OM_cost = pem.calc_opex(electrolyzer_size_mw*1E-3, electrolyzer_total_capital_cost)
        
    #     opex.append(electrolyzer_OM_cost)
    #     capex.append(electrolyzer_total_capital_cost)

    #     for j, div in enumerate(ndivs):

    #         # divided
    #         electrolyzer_size_mw_distributed = electrolyzer_size_mw/div

    #         electrolyzer_capital_cost_distributed = pem.calc_capex(electrolyzer_size_mw_distributed*1E-3, electrolyzer_capex_kw)*1E6
    #         electrolyzer_OM_cost_distributed = pem.calc_opex(electrolyzer_size_mw_distributed*1E-3, electrolyzer_capital_cost_distributed)

    #         # print(opex_distributed)
    #         capex_distributed[j, i] = electrolyzer_capital_cost_distributed*div
    #         opex_distributed[j, i] = electrolyzer_OM_cost_distributed*div
        
    # fig, ax = plt.subplots(1,2, figsize=(6,3))
    # ax[0].plot(electrolyzer_sizes_mw, np.asarray(capex)*1E-6, label="Centralized")
    # ax[1].plot(electrolyzer_sizes_mw, np.asarray(opex)*1E-6, label="Centralized")

    # for i, div in enumerate(ndivs):
    #     # dims(capex_distributed)
    #     ax[0].plot(electrolyzer_sizes_mw, np.asarray(capex_distributed[i])*1E-6, "--", label="%i Divisions" % (div))
    #     ax[1].plot(electrolyzer_sizes_mw, np.asarray(opex_distributed[i])*1E-6, "--", label="%i Divisions" % (div))

    # ax[0].set(ylabel="CAPEX (M USD)", xlabel="Electrolyzer Size (MW)")
    # ax[1].set(ylabel="Annual OPEX (M USD)", xlabel="Electrolyzer Size (MW)")
    # plt.legend(frameon=False)
    # plt.tight_layout()
    # plt.show()

    # ########################### plot divided energy signals
    # fig, ax = plt.subplots(1)
    # ax.plot(electrical_generation_timeseries_kw, label="%s" % (1))
    # for i, div in enumerate(ndivs):
    #     ax.plot(electrical_generation_timeseries_kw/div, label="%s" % (div))

    # ax.set(xlabel="Hour", ylabel="Power (MW)")
    # plt.tight_layout()
    # plt.show()


    ####################### plot installed cost vs reference cost ###############################
    electrolyzer_size_mw = 10
    P_elec =  electrolyzer_size_mw*1E-3 # [GW]
    RC_elec = np.linspace(100, 5000, 1000) # [USD/kW]
    offshore = 0

    pem_offshore = PEMCostsSingliticoModel(elec_location=offshore)

    capex = np.zeros_like(RC_elec)
    opex = np.zeros_like(RC_elec)
    for i, RC in enumerate(RC_elec):
        electrolyzer_capital_cost_musd, electrolyzer_om_cost_musd = pem_offshore.run(P_elec, RC)

        electrolyzer_total_capital_cost = electrolyzer_capital_cost_musd*1E6 # convert from M USD to USD
        electrolyzer_OM_cost = electrolyzer_om_cost_musd*1E6 # convert from M USD to USD
        # print("e tot cap cost: ", electrolyzer_total_capital_cost)
        # print("e tot OM cost: ", electrolyzer_OM_cost)
        capex[i] = electrolyzer_total_capital_cost
        opex[i] = electrolyzer_OM_cost
    
    fig, ax = plt.subplots(2, 1)
    electrolyzer_size_kw = electrolyzer_size_mw*1E3
    ax[0].plot(RC_elec, capex/electrolyzer_size_kw)
    ax[0].set(ylabel="CAPEX ($/kW)")
    ax[1].plot(RC_elec, opex/electrolyzer_size_kw)
    ax[1].set(ylabel="OPEX ($/yr/kW)")
    for axi in ax:
        axi.set(xlabel="Reference Cost ($/kW)")
        axi.grid(visible=True)
    plt.tight_layout()
    plt.show()


import pytest
from pytest import fixture

from greenheart.simulation.technologies.hydrogen.h2_storage.salt_cavern.salt_cavern import SaltCavernStorage

# Test values are based on conclusions of Papadias 2021 and are in 2019 USD
in_dict = {
    "h2_storage_kg": 1000000,
    "system_flow_rate":100000,
    "model":'papadias'}

@fixture
def salt_cavern_storage():
    salt_cavern_storage = SaltCavernStorage(in_dict)

    return salt_cavern_storage


def test_init():
    salt_cavern_storage = SaltCavernStorage(in_dict)

    assert salt_cavern_storage.input_dict is not None
    assert salt_cavern_storage.output_dict is not None

def test_capex_per_kg(salt_cavern_storage):
    salt_cavern_storage_capex_per_kg, _installed_capex, _compressor_capex = salt_cavern_storage.salt_cavern_capex()
    assert salt_cavern_storage_capex_per_kg == pytest.approx(25.18622259358959)

def test_capex(salt_cavern_storage):
    _salt_cavern_storage_capex_per_kg, installed_capex, _compressor_capex = salt_cavern_storage.salt_cavern_capex()
    assert installed_capex == pytest.approx(24992482.4198)

def test_compressor_capex(salt_cavern_storage):
    _salt_cavern_storage_capex_per_kg, _installed_capex, compressor_capex = salt_cavern_storage.salt_cavern_capex()
    assert compressor_capex == pytest.approx(6516166.67163)

def test_capex_output_dict(salt_cavern_storage):
    _salt_caven_storage_capex_per_kg, _installed_capex, _compressor_capex = salt_cavern_storage.salt_cavern_capex()
    assert salt_cavern_storage.output_dict["salt_cavern_storage_capex"] == pytest.approx(24992482.4198)


def test_opex(salt_cavern_storage):
    _salt_cavern_storage_capex_per_kg, _installed_capex, _compressor_capex = salt_cavern_storage.salt_cavern_capex()
    salt_cavern_storage.salt_cavern_opex()
    assert salt_cavern_storage.output_dict["salt_cavern_storage_opex"] == pytest.approx(1461664)


import os
import sys

sys.path.append("")
# from dotenv import load_dotenv
import pandas as pd

from greenheart.simulation.technologies.hydrogen.electrolysis.PEM_H2_LT_electrolyzer_Clusters import PEM_H2_Clusters as PEMClusters
# from PEM_H2_LT_electrolyzer_Clusters import (
#     PEM_H2_Clusters as PEMClusters,
# )

import numpy as np
from numpy import savetxt  # ESG
import matplotlib.pyplot as plt
from mpl_toolkits import mplot3d
import warnings
import math
import scipy
import time
from scipy import interpolate
from pyomo.environ import *
import random
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd


from greenheart.simulation.technologies.hydrogen.electrolysis.optimization_utils_linear import optimize
import time

# from PyOMO import ipOpt !! FOR SANJANA!!
warnings.filterwarnings("ignore")

"""
Perform a LCOH analysis for an offshore wind + Hydrogen PEM system

1. Offshore wind site locations and cost details (4 sites, $1300/kw capex + BOS cost which will come from Orbit Runs)~
2. Cost Scaling Based on Year (Have Weiser et. al report with cost scaling for fixed and floating tech, will implement)
3. Cost Scaling Based on Plant Size (Shields et. Al report)
4. Future Model Development Required:
- Floating Electrolyzer Platform
"""
#
# ---------------------------
#
class run_PEM_clusters:
    """Inputs:
    `electrical_power_signal`: plant power signal in kWh
    `system_size_mw`: total installed electrolyzer capacity (for green steel this is 1000 MW)
    `num_clusters`: number of PEM clusters that can be run independently
    ->ESG note: I have been using num_clusters = 8 for centralized cases
    Nomenclature:
    `cluster`: cluster is built up of 1MW stacks
    `stack`: must be 1MW (because of current PEM model)
    """

    def __init__(
        self,
        electrical_power_signal,
        system_size_mw,
        num_clusters,
        electrolyzer_direct_cost_kw,
        useful_life,
        user_defined_electrolyzer_params,
        verbose=True
    ):
        # nomen
        self.cluster_cap_mw = np.round(system_size_mw / num_clusters)
        # capacity of each cluster, must be a multiple of 1 MW
        
        self.num_clusters = num_clusters
        self.user_params = user_defined_electrolyzer_params
        self.plant_life_yrs = useful_life
        # Do not modify stack_rating_kw or stack_min_power_kw
        # these represent the hard-coded and unmodifiable
        # PEM model basecode
        turndown_ratio=user_defined_electrolyzer_params['turndown_ratio']
        self.stack_rating_kw = 1000  # single stack rating - DO NOT CHANGE
        self.stack_min_power_kw = turndown_ratio * self.stack_rating_kw
        # self.stack_min_power_kw = 0.1 * self.stack_rating_kw
        self.input_power_kw = electrical_power_signal
        self.cluster_min_power = self.stack_min_power_kw * self.cluster_cap_mw
        self.cluster_max_power = self.stack_rating_kw * self.cluster_cap_mw

        # For the optimization problem:
        self.T = len(self.input_power_kw)
        self.farm_power = 1e9
        self.switching_cost = (electrolyzer_direct_cost_kw*0.15*self.cluster_cap_mw * 1000)*(1.48e-4)/(0.26586)
        self.verbose=verbose

    def run_grid_connected_pem(self,system_size_mw,hydrogen_production_capacity_required_kgphr):
        pem=PEMClusters(
                    system_size_mw,
                    self.plant_life_yrs,
                    **self.user_params,
                )

        power_timeseries,stack_current=pem.grid_connected_func(hydrogen_production_capacity_required_kgphr)
        h2_ts, h2_tot =pem.run_grid_connected_workaround(power_timeseries,stack_current)
        #h2_ts, h2_tot = pem.run(power_timeseries)
        h2_df_ts=pd.Series(h2_ts,name='Cluster #0')
        h2_df_tot=pd.Series(h2_tot,name='Cluster #0')
        # h2_df_ts = pd.DataFrame(h2_ts, index=list(h2_ts.keys()), columns=['Cluster #0'])
        # h2_df_tot = pd.DataFrame(h2_tot, index=list(h2_tot.keys()), columns=['Cluster #0'])
        []
        return pd.DataFrame(h2_df_ts),pd.DataFrame(h2_df_tot)



    def run(self, optimize=False):
        # TODO: add control type as input!
        clusters = self.create_clusters()  # initialize clusters
        if optimize:
            power_to_clusters = self.optimize_power_split()  # run Sanjana's code
        else:
            power_to_clusters = self.even_split_power()
        h2_df_ts = pd.DataFrame()
        h2_df_tot = pd.DataFrame()

        col_names = []
        start = time.perf_counter()
        for ci, cluster in enumerate(clusters):
            cl_name = "Cluster #{}".format(ci)
            col_names.append(cl_name)
            h2_ts, h2_tot = clusters[ci].run(power_to_clusters[ci])
            # h2_dict_ts['Cluster #{}'.format(ci)] = h2_ts

            h2_ts_temp = pd.Series(h2_ts, name=cl_name)
            h2_tot_temp = pd.Series(h2_tot, name=cl_name)
            if len(h2_df_tot) == 0:
                # h2_df_ts=pd.concat([h2_df_ts,h2_ts_temp],axis=0,ignore_index=False)
                h2_df_tot = pd.concat(
                    [h2_df_tot, h2_tot_temp], axis=0, ignore_index=False
                )
                h2_df_tot.columns = col_names

                h2_df_ts = pd.concat([h2_df_ts, h2_ts_temp], axis=0, ignore_index=False)
                h2_df_ts.columns = col_names
            else:
                # h2_df_ts = h2_df_ts.join(h2_ts_temp)
                h2_df_tot = h2_df_tot.join(h2_tot_temp)
                h2_df_tot.columns = col_names

                h2_df_ts = h2_df_ts.join(h2_ts_temp)
                h2_df_ts.columns = col_names

        end = time.perf_counter()
        self.clusters = clusters
        if self.verbose:
            print("Took {} sec to run the RUN function".format(round(end - start, 3)))
        return h2_df_ts, h2_df_tot
        # return h2_dict_ts, h2_df_tot

    def optimize_power_split(self):
        number_of_stacks = self.num_clusters
        rated_power = self.cluster_cap_mw * 1000
        tf = 96
        n_times_to_run = int(np.ceil(self.T / tf))
        df = pd.DataFrame({"Wind + PV Generation": self.input_power_kw})
        P_ = None
        I_ = None
        Tr_ = None
        AC = 1
        F_tot = 1
        diff = 0

        for start_time in range(n_times_to_run):
            print(
                f"Optimizing {number_of_stacks} stacks tarting {start_time*tf}hr/{self.T}hr"
            )
            if start_time == 0:

                df["Wind + PV Generation"].replace(0, np.NaN, inplace=True)
                df = df.interpolate()

            P_wind_t = df["Wind + PV Generation"][
                (start_time * tf) : ((start_time * tf) + tf)
            ].values
            start = time.time()
            if P_ is not None:
                P_ = P_[: len(P_wind_t), :]
                I_ = I_[: len(P_wind_t), :]
                Tr_ = Tr_[: len(P_wind_t), :]
            P_tot_opt, P_, H2f, I_, Tr_, P_wind_t, AC, F_tot = optimize(
                P_wind_t,
                T=(len(P_wind_t)),
                n_stacks=(number_of_stacks),
                c_wp=0,
                c_sw=self.switching_cost,
                rated_power=rated_power,
                P_init=P_,
                I_init=I_,
                T_init=Tr_,
                AC_init=AC,
                F_tot_init=F_tot,
            )

            diff += time.time() - start
            if type(AC).__module__ != "numpy":
                AC = np.array(AC)
                F_tot = np.array(F_tot)
            if start_time == 0:
                P_tot_opt_full = P_tot_opt
                P_full = P_
                P_wind_t_full = P_wind_t
                H2f_full = H2f
                I_full = I_
                Tr_full = np.sum(Tr_, axis=0)
                AC_full = AC
                F_tot_full = F_tot

            else:

                P_full = np.vstack((P_full, P_))
                H2f_full = np.vstack((H2f_full, H2f))
                I_full = np.vstack((I_full, I_))
                Tr_full = np.vstack((Tr_full, np.sum(Tr_, axis=0)))
                AC_full = np.vstack((AC_full, (AC)))
                F_tot_full = np.vstack((F_tot_full, (F_tot)))

        return np.transpose(P_full)

    def even_split_power(self):
        start = time.perf_counter()
        # determine how much power to give each cluster
        num_clusters_on = np.floor(self.input_power_kw / self.cluster_min_power)
        num_clusters_on = np.where(
            num_clusters_on > self.num_clusters, self.num_clusters, num_clusters_on
        )
        power_per_cluster = [
            self.input_power_kw[ti] / num_clusters_on[ti]
            if num_clusters_on[ti] > 0
            else 0
            for ti, pwr in enumerate(self.input_power_kw)
        ]

        power_per_to_active_clusters = np.array(power_per_cluster)
        power_to_clusters = np.zeros((len(self.input_power_kw), self.num_clusters))
        for i, cluster_power in enumerate(
            power_per_to_active_clusters
        ):  # np.arange(0,self.n_stacks,1):
            clusters_off = self.num_clusters - int(num_clusters_on[i])
            no_power = np.zeros(clusters_off)
            with_power = cluster_power * np.ones(int(num_clusters_on[i]))
            tot_power = np.concatenate((with_power, no_power))
            power_to_clusters[i] = tot_power

        # power_to_clusters = np.repeat([power_per_cluster],self.num_clusters,axis=0)
        end = time.perf_counter()
        
        if self.verbose:
            print(
                "Took {} sec to run even_split_power function".format(
                    round(end - start, 3)
                )
            )
        # rows are power, columns are stacks [300 x n_stacks]

        return np.transpose(power_to_clusters)

    def max_h2_cntrl(self):
        # run as many at lower power as possible
        []

    def min_deg_cntrl(self):
        # run as few as possible
        []

    def create_clusters(self):
        start = time.perf_counter()
        stacks = []
        # TODO fix the power input - don't make it required!
        # in_dict={'dt':3600}
        for i in range(self.num_clusters):
            # stacks.append(PEMClusters(cluster_size_mw = self.cluster_cap_mw))
            stacks.append(
                PEMClusters(
                    self.cluster_cap_mw,
                    self.plant_life_yrs,
                    **self.user_params,
                )
            )
        end = time.perf_counter()
        if self.verbose:
            print("Took {} sec to run the create clusters".format(round(end - start, 3)))
        return stacks


if __name__ == "__main__":

    system_size_mw = 1000
    num_clusters = 20
    cluster_cap_mw = system_size_mw / num_clusters
    stack_rating_kw = 1000
    cluster_min_power_kw = 0.1 * stack_rating_kw * cluster_cap_mw
    num_steps = 200
    power_rampup = np.arange(
        cluster_min_power_kw, system_size_mw * stack_rating_kw, cluster_min_power_kw
    )

    plant_life = 30
    electrolyzer_model_parameters = {
        "eol_eff_percent_loss":10,
        "uptime_hours_until_eol": 77600,
        "include_degradation_penalty": True,
        "turndown_ratio":0.1,
    }
    # power_rampup = np.linspace(cluster_min_power_kw,system_size_mw*1000,num_steps)
    power_rampdown = np.flip(power_rampup)
    power_in = np.concatenate((power_rampup, power_rampdown))
    pem = run_PEM_clusters(
        power_in,
        system_size_mw,
        num_clusters,
        plant_life,
        electrolyzer_model_parameters,
    )

    h2_ts, h2_tot = pem.run()
    # pem.clusters[0].cell_design(80,1920*2)
    []



# general imports
import os
from typing import Optional, Union
import warnings
import numpy as np
import pandas as pd
from attrs import define, field
import copy

pd.options.mode.chained_assignment = None  # default='warn'

from hopp.simulation import HoppInterface
from ProFAST import ProFAST

from greenheart.simulation.technologies.ammonia.ammonia import (
    run_ammonia_full_model,
    AmmoniaCostModelOutputs,
    AmmoniaFinanceModelOutputs,
    AmmoniaCapacityModelOutputs,
)
from greenheart.simulation.technologies.steel.steel import (
    run_steel_full_model,
    SteelCostModelOutputs,
    SteelFinanceModelOutputs,
    SteelCapacityModelOutputs,
)

# visualization imports
import matplotlib.pyplot as plt

# HOPP imports
import greenheart.tools.eco.electrolysis as he_elec
import greenheart.tools.eco.finance as he_fin
import greenheart.tools.eco.hopp_mgmt as he_hopp
import greenheart.tools.eco.utilities as he_util
import greenheart.tools.eco.hydrogen_mgmt as he_h2


@define
class GreenHeartSimulationConfig:
    """
    Class to hold all the configuration parameters for the GreenHeart model

    Also sets up the HOPP, GreenHeart, ORBIT, and FLORIS configurations based on the
    input files and configuration parameters passed in.

    Args:
        filename_hopp_config (str): filename for the HOPP configuration
        filename_config.greenheart_config (str): filename for the GreenHeart configuration
        filename_turbine_config (str): filename for the turbine configuration
        filename_orbit_config (str): filename for the ORBIT configuration
        filename_floris_config (str): filename for the FLORIS configuration
        electrolyzer_rating_mw (Optional[float]): rating of the electrolyzer in MW
        solar_rating (Optional[float]): rating of the solar plant in MW
        battery_capacity_kw (Optional[float]): capacity of the battery in kW
        battery_capacity_kwh (Optional[float]): capacity of the battery in kWh
        wind_rating (Optional[float]): rating of the wind plant in MW
        verbose (bool): flag to print verbose output
        show_plots (bool): flag to show plots
        save_plots (bool): flag to save plots
        output_dir (str, path): path for saving output files
        use_profast (bool): flag to use profast
        post_processing (bool): flag to run post processing
        storage_type (Optional[str]): type of storage
        incentive_option (int): incentive option
        plant_design_scenario (int): plant design scenario
        output_level (int): output level
        grid_connection (Optional[bool]): flag for grid connection
    """

    filename_hopp_config: str
    filename_greenheart_config: str
    filename_turbine_config: str
    filename_floris_config: str
    filename_orbit_config: Optional[str] = field(default=None)
    electrolyzer_rating_mw: Optional[float] = field(default=None)
    solar_rating: Optional[float] = field(default=None)
    battery_capacity_kw: Optional[float] = field(default=None)
    battery_capacity_kwh: Optional[float] = field(default=None)
    wind_rating: Optional[float] = field(default=None)
    verbose: bool = field(default=False)
    show_plots: bool = field(default=False)
    save_plots: bool = field(default=False)
    output_dir: Optional[Union[str, os.PathLike]] = field(default="output/")
    use_profast: bool = field(default=True)
    post_processing: bool = field(default=True)
    storage_type: Optional[str] = field(default=None)
    incentive_option: int = field(default=1)
    plant_design_scenario: int = field(default=1)
    output_level: int = field(default=8)
    grid_connection: Optional[bool] = field(default=None)

    # these are set in the __attrs_post_init__ method
    hopp_config: dict = field(init=False)
    greenheart_config: dict = field(init=False)
    orbit_config: dict = field(init=False)
    turbine_config: dict = field(init=False)
    floris_config: Optional[dict] = field(init=False)
    orbit_hybrid_electrical_export_config: dict = field(init=False)
    design_scenario: dict = field(init=False)

    def __attrs_post_init__(self):
        (
            self.hopp_config,
            self.greenheart_config,
            self.orbit_config,
            self.turbine_config,
            self.floris_config,
            self.orbit_hybrid_electrical_export_config,
        ) = he_util.get_inputs(
            self.filename_hopp_config,
            self.filename_greenheart_config,
            filename_orbit_config=self.filename_orbit_config,
            filename_floris_config=self.filename_floris_config,
            filename_turbine_config=self.filename_turbine_config,
            verbose=self.verbose,
            show_plots=self.show_plots,
            save_plots=self.save_plots,
        )

        # n scenarios, n discrete variables
        self.design_scenario = self.greenheart_config["plant_design"][
            "scenario%s" % (self.plant_design_scenario)
        ]
        self.design_scenario["id"] = self.plant_design_scenario

        # if design_scenario["h2_storage_location"] == "turbine":
        #     plant_config["h2_storage"]["type"] = "turbine"

        if self.electrolyzer_rating_mw != None:
            self.greenheart_config["electrolyzer"]["flag"] = True
            self.greenheart_config["electrolyzer"][
                "rating"
            ] = self.electrolyzer_rating_mw

        if self.solar_rating != None:
            self.hopp_config["site"]["solar"] = True
            self.hopp_config["technologies"]["pv"][
                "system_capacity_kw"
            ] = self.solar_rating

        if self.battery_capacity_kw != None:
            self.hopp_config["site"]["battery"]["flag"] = True
            self.hopp_config["technologies"]["battery"][
                "system_capacity_kw"
            ] = self.battery_capacity_kw

        if self.battery_capacity_kwh != None:
            self.hopp_config["site"]["battery"]["flag"] = True
            self.hopp_config["technologies"]["battery"][
                "system_capacity_kwh"
            ] = self.battery_capacity_kwh

        if self.storage_type != None:
            self.greenheart_config["h2_storage"]["type"] = self.storage_type

        if self.wind_rating != None:
            self.orbit_config["plant"]["capacity"] = int(self.wind_rating * 1e-3)
            self.orbit_config["plant"]["num_turbines"] = int(
                self.wind_rating * 1e-3 / self.turbine_config["turbine_rating"]
            )
            self.hopp_config["technologies"]["wind"]["num_turbines"] = (
                self.orbit_config["plant"]["num_turbines"]
            )

        if self.grid_connection != None:
            self.greenheart_config["project_parameters"][
                "grid_connection"
            ] = self.grid_connection
            if self.grid_connection:
                self.hopp_config["technologies"]["grid"]["interconnect_kw"] = (
                    self.orbit_config["plant"]["capacity"] * 1e6
                )


@define
class GreenHeartSimulationOutput:
    """This is a dataclass to contain the outputs from GreenHEART

    Args:
        greenheart_config (GreenHeartSimulationConfig): all inputs to the greenheart simulation
        hopp_interface (HoppInterface): the hopp interface created and used by GreenHEART in the simulation
        profast_lcoe (ProFAST): the profast instance used for the lcoe calculations
        profast_lcoh (ProFAST): the profast instance used for the lcoh calculations
        profast_lcoh (ProFAST): the profast instance used for the lcoh calculations if  hydrogen were produced only from the grid
        lcoe (float): levelized cost of energy (electricity)
        lcoh (float): levelized cost of hydrogen
        lcoh_grid_only (float): levelized cost of hydrogen if produced only from the grid
        hopp_results (dict): results from the hopp simulation
        electrolyzer_physics_results (dict): results of the electrolysis simulation
        capex_breakdown (dict): overnight capex broken down by technology
        opex_breakdown_annual (dict): annual operational expenditures broken down by technology
        annual_energy_breakdown (dict): annual energy generation and usage broken down by technology
        hourly_energy_breakdown (dict): hourly energy generation and usage broken down by technology
        remaining_power_profile (np.ndarray): unused power (hourly)
        steel_capacity (Optional[SteelCapacityModelOutputs]): steel capacity information
        steel_costs (Optional[SteelCostModelOutputs]): steel cost information
        steel_finance (Optional[SteelFinanceModelOutputs]): steel financial information
        ammonia_capacity (Optional[AmmoniaCapacityModelOutputs]): ammonia capacity information
        ammonia_costs (Optional[AmmoniaCostModelOutputs]): ammonia cost information
        ammonia_finance (Optional[AmmoniaFinanceModelOutputs]): ammonia finance information
        platform_results (Optional[dict]): equipment platform information/outputs if used
    """

    # detailed simulation information
    greenheart_config: GreenHeartSimulationConfig
    hopp_interface: HoppInterface

    # detailed financial outputs
    profast_lcoe: ProFAST
    profast_lcoh: ProFAST
    profast_lcoh_grid_only: ProFAST

    # high-level results
    lcoe: float
    lcoh: float
    lcoh_grid_only: float

    # detailed output information
    hopp_results: dict
    electrolyzer_physics_results: dict
    capex_breakdown: dict
    opex_breakdown_annual: dict
    annual_energy_breakdown: dict
    hourly_energy_breakdown: dict
    remaining_power_profile: np.ndarray

    # optional outputs
    h2_storage_max_fill_rate_kg_hr: Optional[dict] = field(default=None)
    h2_storage_capacity_kg: Optional[dict] = field(default=None)
    hydrogen_storage_state_of_charge_kg: Optional[dict] = field(default=None)

    steel_capacity: Optional[SteelCapacityModelOutputs] = field(default=None)
    steel_costs: Optional[SteelCostModelOutputs] = field(default=None)
    steel_finance: Optional[SteelFinanceModelOutputs] = field(default=None)

    ammonia_capacity: Optional[AmmoniaCapacityModelOutputs] = field(default=None)
    ammonia_costs: Optional[AmmoniaCostModelOutputs] = field(default=None)
    ammonia_finance: Optional[AmmoniaFinanceModelOutputs] = field(default=None)

    platform_results: Optional[dict] = field(default=None)


def setup_greenheart_simulation(config: GreenHeartSimulationConfig):

    # run orbit for wind plant construction and other costs
    ## TODO get correct weather (wind, wave) inputs for ORBIT input (possibly via ERA5)
    if config.design_scenario["wind_location"] == "offshore":

        if (
            config.orbit_config["plant"]["num_turbines"]
            != config.hopp_config["technologies"]["wind"]["num_turbines"]
        ):
            config.orbit_config["plant"].update(
                {
                    "num_turbines": config.hopp_config["technologies"]["wind"][
                        "num_turbines"
                    ]
                }
            )
            warnings.warn(
                f"'num_turbines' in the orbit_config was {config.orbit_config['plant']['num_turbines']}, but 'num_turbines' in"
                f"hopp_config was {config.hopp_config['technologies']['wind']['num_turbines']}. The 'num_turbines' value in the orbit_config"
                "is being overwritten with the value from the hopp_config",
                UserWarning,
            )

        if (
            config.orbit_config["site"]["depth"]
            != config.greenheart_config["site"]["depth"]
        ):
            config.orbit_config["site"].update(
                {"depth": config.greenheart_config["site"]["depth"]}
            )
            warnings.warn(
                f"site depth in the orbit_config was {config.orbit_config['site']['depth']}, but site depth in"
                f"greenheart_config was {config.greenheart_config['site']['depth']}. The site depth value in the orbit_config"
                "is being overwritten with the value from the greenheart_config",
                UserWarning,
            )

        if (
            config.orbit_config["plant"]["turbine_spacing"]
            != config.greenheart_config["site"]["wind_layout"]["turbine_spacing"]
        ):
            config.orbit_config["plant"].update(
                {
                    "turbine_spacing": config.greenheart_config["site"]["wind_layout"][
                        "turbine_spacing"
                    ]
                }
            )
            warnings.warn(
                f"'turbine_spacing' in the orbit_config was {config.orbit_config['plant']['turbine_spacing']}, but 'turbine_spacing' in"
                f"greenheart_config was {config.greenheart_config['site']['wind_layout']['turbine_spacing']}. The 'turbine_spacing' value in the orbit_config"
                "is being overwritten with the value from the greenheart_config",
                UserWarning,
            )

        if (
            config.orbit_config["plant"]["row_spacing"]
            != config.greenheart_config["site"]["wind_layout"]["row_spacing"]
        ):
            config.orbit_config["plant"].update(
                {
                    "row_spacing": config.greenheart_config["site"]["wind_layout"][
                        "row_spacing"
                    ]
                }
            )
            warnings.warn(
                f"'row_spacing' in the orbit_config was {config.orbit_config['plant']['row_spacing']}, but 'row_spacing' in"
                f"greenheart_config was {config.greenheart_config['site']['wind_layout']['row_spacing']}. The 'row_spacing' value in the orbit_config"
                "is being overwritten with the value from the greenheart_config",
                UserWarning,
            )

        wind_config = he_fin.WindCostConfig(
            design_scenario=config.design_scenario,
            hopp_config=config.hopp_config,
            greenheart_config=config.greenheart_config,
            orbit_config=config.orbit_config,
            orbit_hybrid_electrical_export_config=config.orbit_hybrid_electrical_export_config,
        )

        wind_cost_results = he_fin.run_wind_cost_model(
            wind_cost_inputs=wind_config, verbose=config.verbose
        )
    else:
        wind_cost_results = None

    # override individual fin_model values with cost_info values
    if "wind" in config.hopp_config["technologies"]:
        if ("wind_om_per_kw" in config.hopp_config["config"]["cost_info"]) and (
            config.hopp_config["technologies"]["wind"]["fin_model"]["system_costs"][
                "om_capacity"
            ][0]
            != config.hopp_config["config"]["cost_info"]["wind_om_per_kw"]
        ):

            for i in range(
                len(
                    config.hopp_config["technologies"]["wind"]["fin_model"][
                        "system_costs"
                    ]["om_capacity"]
                )
            ):
                config.hopp_config["technologies"]["wind"]["fin_model"]["system_costs"][
                    "om_capacity"
                ][i] = config.hopp_config["config"]["cost_info"]["wind_om_per_kw"]

                om_fixed_wind_fin_model = config.hopp_config["technologies"]["wind"][
                    "fin_model"
                ]["system_costs"]["om_capacity"][i]
                wind_om_per_kw = config.hopp_config["config"]["cost_info"][
                    "wind_om_per_kw"
                ]
                warnings.warn(
                    f"'om_capacity[{i}]' in the wind 'fin_model' was {om_fixed_wind_fin_model}, but 'wind_om_per_kw' in"
                    f"'cost_info' was {wind_om_per_kw}. The 'om_capacity' value in the wind 'fin_model'"
                    "is being overwritten with the value from the 'cost_info'",
                    UserWarning,
                )
        if ("wind_om_per_mwh" in config.hopp_config["config"]["cost_info"]) and (
            config.hopp_config["technologies"]["wind"]["fin_model"]["system_costs"][
                "om_production"
            ][0]
            != config.hopp_config["config"]["cost_info"]["wind_om_per_mwh"]
        ):
            # Use this to set the Production-based O&M amount [$/MWh]
            for i in range(
                len(
                    config.hopp_config["technologies"]["wind"]["fin_model"][
                        "system_costs"
                    ]["om_production"]
                )
            ):
                config.hopp_config["technologies"]["wind"]["fin_model"]["system_costs"][
                    "om_production"
                ][i] = config.hopp_config["config"]["cost_info"]["wind_om_per_mwh"]
            om_wind_variable_cost = config.hopp_config["technologies"]["wind"][
                "fin_model"
            ]["system_costs"]["om_production"][i]
            wind_om_per_mwh = config.hopp_config["config"]["cost_info"][
                "wind_om_per_mwh"
            ]
            warnings.warn(
                f"'om_production' in the wind 'fin_model' was {om_wind_variable_cost}, but 'wind_om_per_mwh' in"
                f"'cost_info' was {wind_om_per_mwh}. The 'om_production' value in the wind 'fin_model'"
                "is being overwritten with the value from the 'cost_info'",
                UserWarning,
            )

    if "pv" in config.hopp_config["technologies"]:
        if ("pv_om_per_kw" in config.hopp_config["config"]["cost_info"]) and (
            config.hopp_config["technologies"]["pv"]["fin_model"]["system_costs"][
                "om_capacity"
            ][0]
            != config.hopp_config["config"]["cost_info"]["pv_om_per_kw"]
        ):
            for i in range(
                len(
                    config.hopp_config["technologies"]["pv"]["fin_model"][
                        "system_costs"
                    ]["om_capacity"]
                )
            ):
                config.hopp_config["technologies"]["pv"]["fin_model"]["system_costs"][
                    "om_capacity"
                ][i] = config.hopp_config["config"]["cost_info"]["pv_om_per_kw"]

                om_fixed_pv_fin_model = config.hopp_config["technologies"]["pv"][
                    "fin_model"
                ]["system_costs"]["om_capacity"][i]
                pv_om_per_kw = config.hopp_config["config"]["cost_info"]["pv_om_per_kw"]
                warnings.warn(
                    f"'om_capacity[{i}]' in the pv 'fin_model' was {om_fixed_pv_fin_model}, but 'pv_om_per_kw' in"
                    f"'cost_info' was {pv_om_per_kw}. The 'om_capacity' value in the pv 'fin_model'"
                    "is being overwritten with the value from the 'cost_info'",
                    UserWarning,
                )
        if ("pv_om_per_mwh" in config.hopp_config["config"]["cost_info"]) and (
            config.hopp_config["technologies"]["pv"]["fin_model"]["system_costs"][
                "om_production"
            ][0]
            != config.hopp_config["config"]["cost_info"]["pv_om_per_mwh"]
        ):
            # Use this to set the Production-based O&M amount [$/MWh]
            for i in range(
                len(
                    config.hopp_config["technologies"]["pv"]["fin_model"][
                        "system_costs"
                    ]["om_production"]
                )
            ):
                config.hopp_config["technologies"]["pv"]["fin_model"]["system_costs"][
                    "om_production"
                ][i] = config.hopp_config["config"]["cost_info"]["pv_om_per_mwh"]
            om_pv_variable_cost = config.hopp_config["technologies"]["pv"]["fin_model"][
                "system_costs"
            ]["om_production"][i]
            pv_om_per_mwh = config.hopp_config["config"]["cost_info"]["pv_om_per_mwh"]
            warnings.warn(
                f"'om_production' in the pv 'fin_model' was {om_pv_variable_cost}, but 'pv_om_per_mwh' in"
                f"'cost_info' was {pv_om_per_mwh}. The 'om_production' value in the pv 'fin_model'"
                "is being overwritten with the value from the 'cost_info'",
                UserWarning,
            )

    if "battery" in config.hopp_config["technologies"]:
        if ("battery_om_per_kw" in config.hopp_config["config"]["cost_info"]) and (
            config.hopp_config["technologies"]["battery"]["fin_model"]["system_costs"][
                "om_capacity"
            ][0]
            != config.hopp_config["config"]["cost_info"]["battery_om_per_kw"]
        ):
            for i in range(
                len(
                    config.hopp_config["technologies"]["battery"]["fin_model"][
                        "system_costs"
                    ]["om_capacity"]
                )
            ):
                config.hopp_config["technologies"]["battery"]["fin_model"][
                    "system_costs"
                ]["om_capacity"][i] = config.hopp_config["config"]["cost_info"][
                    "battery_om_per_kw"
                ]

            om_batt_fixed_cost = config.hopp_config["technologies"]["battery"][
                "fin_model"
            ]["system_costs"]["om_capacity"][i]
            battery_om_per_kw = config.hopp_config["config"]["cost_info"][
                "battery_om_per_kw"
            ]
            warnings.warn(
                f"'om_capacity' in the battery 'fin_model' was {om_batt_fixed_cost}, but 'battery_om_per_kw' in"
                f"'cost_info' was {battery_om_per_kw}. The 'om_capacity' value in the battery 'fin_model'"
                "is being overwritten with the value from the 'cost_info'",
                UserWarning,
            )
        if ("battery_om_per_mwh" in config.hopp_config["config"]["cost_info"]) and (
            config.hopp_config["technologies"]["battery"]["fin_model"]["system_costs"][
                "om_production"
            ][0]
            != config.hopp_config["config"]["cost_info"]["battery_om_per_mwh"]
        ):
            # Use this to set the Production-based O&M amount [$/MWh]
            for i in range(
                len(
                    config.hopp_config["technologies"]["battery"]["fin_model"][
                        "system_costs"
                    ]["om_production"]
                )
            ):
                config.hopp_config["technologies"]["battery"]["fin_model"][
                    "system_costs"
                ]["om_production"][i] = config.hopp_config["config"]["cost_info"][
                    "battery_om_per_mwh"
                ]
            om_batt_variable_cost = config.hopp_config["technologies"]["battery"][
                "fin_model"
            ]["system_costs"]["om_production"][i]
            battery_om_per_mwh = config.hopp_config["config"]["cost_info"][
                "battery_om_per_mwh"
            ]
            warnings.warn(
                f"'om_production' in the battery 'fin_model' was {om_batt_variable_cost}, but 'battery_om_per_mwh' in"
                f"'cost_info' was {battery_om_per_mwh}. The 'om_production' value in the battery 'fin_model'"
                "is being overwritten with the value from the 'cost_info'",
                UserWarning,
            )

    # setup HOPP model
    hi = he_hopp.setup_hopp(
        config.hopp_config,
        config.greenheart_config,
        config.orbit_config,
        config.turbine_config,
        config.floris_config,
        config.design_scenario,
        wind_cost_results,
        show_plots=config.show_plots,
        save_plots=config.save_plots,
    )

    return config, hi, wind_cost_results


def run_simulation(config: GreenHeartSimulationConfig):

    config, hi, wind_cost_results = setup_greenheart_simulation(config=config)

    # run HOPP model
    # hopp_results = he_hopp.run_hopp(hopp_site, hopp_technologies, hopp_scenario, hopp_h2_args, verbose=verbose)
    hopp_results = he_hopp.run_hopp(
        hi,
        project_lifetime=config.greenheart_config["project_parameters"][
            "project_lifetime"
        ],
        verbose=config.verbose,
    )

    if config.design_scenario["wind_location"] == "onshore":
        wind_config = he_fin.WindCostConfig(
            design_scenario=config.design_scenario,
            hopp_config=config.hopp_config,
            greenheart_config=config.greenheart_config,
            turbine_config=config.turbine_config,
            hopp_interface=hopp_results["hopp_interface"],
        )

        wind_cost_results = he_fin.run_wind_cost_model(
            wind_cost_inputs=wind_config, verbose=config.verbose
        )

    # this portion of the system is inside a function so we can use a solver to determine the correct energy availability for h2 production
    def energy_internals(
        hopp_results=hopp_results,
        wind_cost_results=wind_cost_results,
        design_scenario=config.design_scenario,
        orbit_config=config.orbit_config,
        hopp_config=config.hopp_config,
        greenheart_config=config.greenheart_config,
        turbine_config=config.turbine_config,
        wind_resource=hi.system.site.wind_resource,
        verbose=config.verbose,
        show_plots=config.show_plots,
        save_plots=config.save_plots,
        output_dir=config.output_dir,
        solver=True,
        power_for_peripherals_kw_in=0.0,
        breakdown=False,
    ):

        hopp_results_internal = dict(hopp_results)

        # set energy input profile
        ### subtract peripheral power from supply to get what is left for electrolyzer
        remaining_power_profile_in = np.zeros_like(
            hopp_results["combined_hybrid_power_production_hopp"]
        )

        high_count = sum(
            np.asarray(hopp_results["combined_hybrid_power_production_hopp"])
            >= power_for_peripherals_kw_in
        )
        total_peripheral_energy = power_for_peripherals_kw_in * 365 * 24
        distributed_peripheral_power = total_peripheral_energy / high_count

        remaining_power_profile_in = np.where(
            hopp_results["combined_hybrid_power_production_hopp"]
                - distributed_peripheral_power > 0,
            hopp_results["combined_hybrid_power_production_hopp"]
                - distributed_peripheral_power,
            0
        )

        hopp_results_internal["combined_hybrid_power_production_hopp"] = tuple(
            remaining_power_profile_in
        )

        # run electrolyzer physics model
        electrolyzer_physics_results = he_elec.run_electrolyzer_physics(
            hopp_results_internal,
            config.greenheart_config,
            wind_resource,
            design_scenario,
            show_plots=show_plots,
            save_plots=save_plots,
            output_dir=output_dir,
            verbose=verbose,
        )

        # run electrolyzer cost model
        electrolyzer_cost_results = he_elec.run_electrolyzer_cost(
            electrolyzer_physics_results,
            hopp_config,
            config.greenheart_config,
            design_scenario,
            verbose=verbose,
        )

        # run electrolyzer bop model
        electrolyzer_energy_consumption_bop_kw = he_elec.run_electrolyzer_bop(
            greenheart_config,
            electrolyzer_physics_results
        )

        desal_results = he_elec.run_desal(
            hopp_config, electrolyzer_physics_results, design_scenario, verbose
        )

        # run array system model
        h2_pipe_array_results = he_h2.run_h2_pipe_array(
            greenheart_config,
            hopp_config,
            turbine_config,
            wind_cost_results,
            electrolyzer_physics_results,
            design_scenario,
            verbose,
        )

        # compressor #TODO size correctly
        (
            h2_transport_compressor,
            h2_transport_compressor_results,
        ) = he_h2.run_h2_transport_compressor(
            config.greenheart_config,
            electrolyzer_physics_results,
            design_scenario,
            verbose=verbose,
        )

        # transport pipeline
        if design_scenario["wind_location"] == "offshore":
            h2_transport_pipe_results = he_h2.run_h2_transport_pipe(
                orbit_config,
                greenheart_config,
                electrolyzer_physics_results,
                design_scenario,
                verbose=verbose,
            )
        if design_scenario["wind_location"] == "onshore":
            h2_transport_pipe_results = {
                "total capital cost [$]": [0 * 5433290.0184895478],
                "annual operating cost [$]": [0.0],
            }

        # pressure vessel storage
        pipe_storage, h2_storage_results = he_h2.run_h2_storage(
            hopp_config,
            greenheart_config,
            turbine_config,
            electrolyzer_physics_results,
            design_scenario,
            verbose=verbose,
        )

        total_energy_available = np.sum(
            hopp_results["combined_hybrid_power_production_hopp"]
        )

        ### get all energy non-electrolyzer usage in kw
        desal_power_kw = desal_results["power_for_desal_kw"]

        h2_transport_compressor_power_kw = h2_transport_compressor_results[
            "compressor_power"
        ]  # kW

        h2_storage_energy_kwh = h2_storage_results["storage_energy"]
        h2_storage_power_kw = h2_storage_energy_kwh * (1.0 / (365 * 24))

        total_accessory_power_renewable_kw = np.zeros(len(electrolyzer_energy_consumption_bop_kw))
        total_accessory_power_renewable_kw += electrolyzer_energy_consumption_bop_kw
        # if transport is not HVDC and h2 storage is on shore, then power the storage from the grid
        if (design_scenario["transportation"] == "pipeline") and (
            design_scenario["h2_storage_location"] == "onshore"
        ):
            total_accessory_power_renewable_kw += (
                desal_power_kw + h2_transport_compressor_power_kw
            )
            total_accessory_power_grid_kw = h2_storage_power_kw
        else:
            total_accessory_power_renewable_kw += (
                desal_power_kw + h2_transport_compressor_power_kw + h2_storage_power_kw
            )
            total_accessory_power_grid_kw = 0.0

        ### subtract peripheral power from supply to get what is left for electrolyzer and also get grid power
        remaining_power_profile = np.zeros_like(
            hopp_results["combined_hybrid_power_production_hopp"]
        )
        grid_power_profile = np.zeros_like(
            hopp_results["combined_hybrid_power_production_hopp"]
        )
        remaining_power_profile = np.where(
            hopp_results["combined_hybrid_power_production_hopp"]
                - total_accessory_power_renewable_kw > 0,
            hopp_results["combined_hybrid_power_production_hopp"]
                - total_accessory_power_renewable_kw,
            0
        )

        if verbose and not solver:
            print("\nEnergy/Power Results:")
            print("Supply (MWh): ", total_energy_available)
            print("Desal (kW): ", desal_power_kw)
            print("Transport compressor (kW): ", h2_transport_compressor_power_kw)
            print("Storage compression, refrigeration, etc (kW): ", h2_storage_power_kw)
            # print("Difference: ", total_energy_available/(365*24) - np.sum(remaining_power_profile)/(365*24) - total_accessory_power_renewable_kw)

        if (show_plots or save_plots) and not solver:
            fig, ax = plt.subplots(1)
            plt.plot(
                np.asarray(hopp_results["combined_hybrid_power_production_hopp"])
                * 1e-6,
                label="Total Energy Available",
            )
            plt.plot(
                remaining_power_profile * 1e-6,
                label="Energy Available for Electrolysis",
            )
            plt.xlabel("Hour")
            plt.ylabel("Power (GW)")
            plt.tight_layout()
            if save_plots:
                savepath = config.output_dir + "figures/power_series/"
                if not os.path.exists(savepath):
                    os.makedirs(savepath)
                plt.savefig(
                    savepath + "power_%i.png" % (design_scenario["id"]),
                    transparent=True,
                )
            if show_plots:
                plt.show()
        if solver:
            if breakdown:
                return (
                    total_accessory_power_renewable_kw,
                    total_accessory_power_grid_kw,
                    desal_power_kw,
                    h2_transport_compressor_power_kw,
                    h2_storage_power_kw,
                    electrolyzer_energy_consumption_bop_kw,
                    remaining_power_profile,
                )
            else:
                return total_accessory_power_renewable_kw
        else:
            return (
                electrolyzer_physics_results,
                electrolyzer_cost_results,
                desal_results,
                h2_pipe_array_results,
                h2_transport_compressor,
                h2_transport_compressor_results,
                h2_transport_pipe_results,
                pipe_storage,
                h2_storage_results,
                total_accessory_power_renewable_kw,
                total_accessory_power_grid_kw,
                remaining_power_profile,
            )

    # define function to provide to the brent solver
    def energy_residual_function(power_for_peripherals_kw_in):

        # get results for current design
        power_for_peripherals_kw_out = energy_internals(
            power_for_peripherals_kw_in=power_for_peripherals_kw_in,
            solver=True,
            verbose=False,
        )

        # collect residual
        power_residual = power_for_peripherals_kw_out - power_for_peripherals_kw_in

        return power_residual

    def simple_solver(initial_guess=0.0):

        # get results for current design
        (
            total_accessory_power_renewable_kw,
            total_accessory_power_grid_kw,
            desal_power_kw,
            h2_transport_compressor_power_kw,
            h2_storage_power_kw,
            electrolyzer_bop_kw,
            remaining_power_profile,
        ) = energy_internals(
            power_for_peripherals_kw_in=initial_guess,
            solver=True,
            verbose=False,
            breakdown=True,
        )

        return (
            total_accessory_power_renewable_kw,
            total_accessory_power_grid_kw,
            desal_power_kw,
            h2_transport_compressor_power_kw,
            h2_storage_power_kw,
            electrolyzer_bop_kw
        )

    #################### solving for energy needed for non-electrolyzer components ####################################
    # this approach either exactly over over-estimates the energy needed for non-electrolyzer components
    solver_results = simple_solver(0)
    solver_result = solver_results[0]

    # # this is a check on the simple solver
    # print("\nsolver result: ", solver_result)
    # residual = energy_residual_function(solver_result)
    # print("\nresidual: ", residual)

    # this approach exactly sizes the energy needed for the non-electrolyzer components (according to the current models anyway)
    # solver_result = optimize.brentq(energy_residual_function, -10, 20000, rtol=1E-5)
    # OptimizeResult = optimize.root(energy_residual_function, 11E3, tol=1)
    # solver_result = OptimizeResult.x
    # solver_results = simple_solver(solver_result)
    # solver_result = solver_results[0]
    # print(solver_result)

    ##################################################################################################################

    # get results for final design
    (
        electrolyzer_physics_results,
        electrolyzer_cost_results,
        desal_results,
        h2_pipe_array_results,
        h2_transport_compressor,
        h2_transport_compressor_results,
        h2_transport_pipe_results,
        pipe_storage,
        h2_storage_results,
        total_accessory_power_renewable_kw,
        total_accessory_power_grid_kw,
        remaining_power_profile,
    ) = energy_internals(solver=False, power_for_peripherals_kw_in=solver_result)

    ## end solver loop here
    platform_results = he_h2.run_equipment_platform(
        config.hopp_config,
        config.greenheart_config,
        config.orbit_config,
        config.design_scenario,
        hopp_results,
        electrolyzer_physics_results,
        h2_storage_results,
        desal_results,
        verbose=config.verbose,
    )

    ################# OSW intermediate calculations" aka final financial calculations
    # does LCOE even make sense if we are only selling the H2? I think in this case LCOE should not be used, rather LCOH should be used. Or, we could use LCOE based on the electricity actually used for h2
    # I think LCOE is just being used to estimate the cost of the electricity used, but in this case we should just use the cost of the electricity generating plant since we are not selling to the grid. We
    # could build in a grid connection later such that we use LCOE for any purchased electricity and sell any excess electricity after H2 production
    # actually, I think this is what OSW is doing for LCOH

    # TODO double check full-system CAPEX
    capex, capex_breakdown = he_fin.run_capex(
        hopp_results,
        wind_cost_results,
        electrolyzer_cost_results,
        h2_pipe_array_results,
        h2_transport_compressor_results,
        h2_transport_pipe_results,
        h2_storage_results,
        config.hopp_config,
        config.greenheart_config,
        config.design_scenario,
        desal_results,
        platform_results,
        verbose=config.verbose,
    )

    # TODO double check full-system OPEX
    opex_annual, opex_breakdown_annual = he_fin.run_opex(
        hopp_results,
        wind_cost_results,
        electrolyzer_cost_results,
        h2_pipe_array_results,
        h2_transport_compressor_results,
        h2_transport_pipe_results,
        h2_storage_results,
        config.hopp_config,
        config.greenheart_config,
        desal_results,
        platform_results,
        verbose=config.verbose,
        total_export_system_cost=capex_breakdown["electrical_export_system"],
    )

    if config.verbose:
        print(
            "hybrid plant capacity factor: ",
            np.sum(hopp_results["combined_hybrid_power_production_hopp"])
            / (hopp_results["hybrid_plant"].system_capacity_kw.hybrid * 365 * 24),
        )

    steel_finance = None
    ammonia_finance = None

    if config.use_profast:
        lcoe, pf_lcoe = he_fin.run_profast_lcoe(
            config.greenheart_config,
            wind_cost_results,
            capex_breakdown,
            opex_breakdown_annual,
            hopp_results,
            config.incentive_option,
            config.design_scenario,
            verbose=config.verbose,
            show_plots=config.show_plots,
            save_plots=config.save_plots,
            output_dir=config.output_dir,
        )
        lcoh_grid_only, pf_grid_only = he_fin.run_profast_grid_only(
            config.greenheart_config,
            wind_cost_results,
            electrolyzer_physics_results,
            capex_breakdown,
            opex_breakdown_annual,
            hopp_results,
            config.design_scenario,
            total_accessory_power_renewable_kw,
            total_accessory_power_grid_kw,
            verbose=config.verbose,
            show_plots=config.show_plots,
            save_plots=config.save_plots,
            output_dir=config.output_dir,
        )
        lcoh, pf_lcoh = he_fin.run_profast_full_plant_model(
            config.greenheart_config,
            wind_cost_results,
            electrolyzer_physics_results,
            capex_breakdown,
            opex_breakdown_annual,
            hopp_results,
            config.incentive_option,
            config.design_scenario,
            total_accessory_power_renewable_kw,
            total_accessory_power_grid_kw,
            verbose=config.verbose,
            show_plots=config.show_plots,
            save_plots=config.save_plots,
            output_dir=config.output_dir,
        )

        hydrogen_amount_kgpy = electrolyzer_physics_results["H2_Results"][
            "Life: Annual H2 production [kg/year]"
        ]

        if "steel" in config.greenheart_config:
            steel_config = copy.deepcopy(config.greenheart_config)
            if config.verbose:
                print("Running steel\n")

            # use lcoh from the electrolyzer model if it is not already in the config
            if "lcoh" not in steel_config["steel"]["finances"]:
                steel_config["steel"]["finances"]["lcoh"] = lcoh

            # use lcoh from the electrolyzer model if it is not already in the config
            if "lcoh" not in steel_config["steel"]["costs"]:
                steel_config["steel"]["costs"]["lcoh"] = lcoh

            # use the hydrogen amount from the electrolyzer physics model if it is not already in the config
            if "hydrogen_amount_kgpy" not in steel_config["steel"]["capacity"]:
                steel_config["steel"]["capacity"][
                    "hydrogen_amount_kgpy"
                ] = hydrogen_amount_kgpy

            steel_capacity, steel_costs, steel_finance = run_steel_full_model(
                steel_config,
                save_plots=config.save_plots,
                show_plots=config.show_plots,
                output_dir=config.output_dir,
                design_scenario_id=config.design_scenario["id"],
            )

        else:
            steel_finance = {}

        if "ammonia" in config.greenheart_config:
            ammonia_config = copy.deepcopy(config.greenheart_config)
            if config.verbose:
                print("Running ammonia\n")

            if "hydrogen_cost" not in ammonia_config["ammonia"]["costs"]["feedstocks"]:
                ammonia_config["ammonia"]["costs"]["feedstocks"]["hydrogen_cost"] = lcoh

            # use the hydrogen amount from the electrolyzer physics model if it is not already in the config
            if "hydrogen_amount_kgpy" not in ammonia_config["ammonia"]["capacity"]:
                ammonia_config["ammonia"]["capacity"][
                    "hydrogen_amount_kgpy"
                ] = hydrogen_amount_kgpy

            ammonia_capacity, ammonia_costs, ammonia_finance = run_ammonia_full_model(
                ammonia_config,
                save_plots=config.save_plots,
                show_plots=config.show_plots,
                output_dir=config.output_dir,
                design_scenario_id=config.design_scenario["id"],
            )

        else:
            ammonia_finance = {}

    ################# end OSW intermediate calculations
    if config.post_processing:
        annual_energy_breakdown, hourly_energy_breakdown = (
            he_util.post_process_simulation(
                lcoe,
                lcoh,
                pf_lcoh,
                pf_lcoe,
                hopp_results,
                electrolyzer_physics_results,
                config.hopp_config,
                config.greenheart_config,
                config.orbit_config,
                config.turbine_config,
                h2_storage_results,
                capex_breakdown,
                opex_breakdown_annual,
                wind_cost_results,
                platform_results,
                desal_results,
                config.design_scenario,
                config.plant_design_scenario,
                config.incentive_option,
                solver_results=solver_results,
                show_plots=config.show_plots,
                save_plots=config.save_plots,
                verbose=config.verbose,
                output_dir=config.output_dir,
            )
        )  # , lcoe, lcoh, lcoh_with_grid, lcoh_grid_only)

    # return
    if config.output_level == 0:
        return 0
    elif config.output_level == 1:
        return lcoh
    elif config.output_level == 2:
        return (
            lcoh,
            lcoe,
            capex_breakdown,
            opex_breakdown_annual,
            pf_lcoh,
            electrolyzer_physics_results,
        )
    elif config.output_level == 3:
        return (
            lcoh,
            lcoe,
            capex_breakdown,
            opex_breakdown_annual,
            pf_lcoh,
            electrolyzer_physics_results,
            pf_lcoe,
            annual_energy_breakdown,
        )
    elif config.output_level == 4:
        return lcoe, lcoh, lcoh_grid_only
    elif config.output_level == 5:
        return lcoe, lcoh, lcoh_grid_only, hopp_results["hopp_interface"]
    elif config.output_level == 6:
        return hopp_results, electrolyzer_physics_results, remaining_power_profile
    elif config.output_level == 7:
        return lcoe, lcoh, steel_finance, ammonia_finance
    elif config.output_level == 8:
        return GreenHeartSimulationOutput(
            config,
            hi,
            pf_lcoe,
            pf_lcoh,
            pf_grid_only,
            lcoe,
            lcoh,
            lcoh_grid_only,
            hopp_results,
            electrolyzer_physics_results,
            capex_breakdown,
            opex_breakdown_annual,
            annual_energy_breakdown,
            hourly_energy_breakdown,
            remaining_power_profile,
            h2_storage_max_fill_rate_kg_hr=(
                None
                if "h2_storage_max_fill_rate_kg_hr" not in h2_storage_results
                else h2_storage_results["h2_storage_max_fill_rate_kg_hr"]
            ),
            h2_storage_capacity_kg=(
                None
                if "h2_storage_capacity_kg" not in h2_storage_results
                else h2_storage_results["h2_storage_capacity_kg"]
            ),
            hydrogen_storage_state_of_charge_kg=(
                None
                if "hydrogen_storage_soc" not in h2_storage_results
                else h2_storage_results["hydrogen_storage_soc"]
            ),
            steel_capacity=(
                None if "steel" not in config.greenheart_config else steel_capacity
            ),
            steel_costs=(
                None if "steel" not in config.greenheart_config else steel_costs
            ),
            steel_finance=(
                None if "steel" not in config.greenheart_config else steel_finance
            ),
            ammonia_capacity=(
                None if "ammonia" not in config.greenheart_config else ammonia_capacity
            ),
            ammonia_costs=(
                None if "ammonia" not in config.greenheart_config else ammonia_costs
            ),
            ammonia_finance=(
                None if "ammonia" not in config.greenheart_config else ammonia_finance
            ),
            platform_results=platform_results,
        )


def run_sweeps(
    simulate=False,
    verbose=True,
    show_plots=True,
    use_profast=True,
    output_dir="output/",
):

    if simulate:
        verbose = False
        show_plots = False
    if simulate:
        storage_types = ["none", "pressure_vessel", "pipe", "salt_cavern"]
        wind_ratings = [400]  # , 800, 1200] #[200, 400, 600, 800]

        for wind_rating in wind_ratings:
            ratings = np.linspace(
                round(0.2 * wind_rating, ndigits=0), 2 * wind_rating + 1, 50
            )
            for storage_type in storage_types:
                lcoh_array = np.zeros(len(ratings))
                for z in np.arange(0, len(ratings)):
                    lcoh_array[z] = run_simulation(
                        electrolyzer_rating_mw=ratings[z],
                        wind_rating=wind_rating,
                        verbose=verbose,
                        show_plots=show_plots,
                        use_profast=use_profast,
                        storage_type=storage_type,
                    )
                    print(lcoh_array)
                np.savetxt(
                    output_dir
                    + "data/lcoh_vs_rating_%s_storage_%sMWwindplant.txt"
                    % (storage_type, wind_rating),
                    np.c_[ratings, lcoh_array],
                )

    if show_plots:

        wind_ratings = [400, 800, 1200]  # [200, 400, 600, 800]
        indexes = [(0, 0), (0, 1), (1, 0), (1, 1)]
        fig, ax = plt.subplots(2, 2, sharex=True, sharey=True, figsize=(10, 6))

        for i in np.arange(0, len(wind_ratings)):
            wind_rating = wind_ratings[i]
            data_no_storage = np.loadtxt(
                "data/lcoh_vs_rating_none_storage_%sMWwindplant.txt" % (wind_rating)
            )
            data_pressure_vessel = np.loadtxt(
                "data/lcoh_vs_rating_pressure_vessel_storage_%sMWwindplant.txt"
                % (wind_rating)
            )
            data_salt_cavern = np.loadtxt(
                "data/lcoh_vs_rating_salt_cavern_storage_%sMWwindplant.txt"
                % (wind_rating)
            )
            data_pipe = np.loadtxt(
                "data/lcoh_vs_rating_pipe_storage_%sMWwindplant.txt" % (wind_rating)
            )

            ax[indexes[i]].plot(
                data_pressure_vessel[:, 0] / wind_rating,
                data_pressure_vessel[:, 1],
                label="Pressure Vessel",
            )
            ax[indexes[i]].plot(
                data_pipe[:, 0] / wind_rating, data_pipe[:, 1], label="Underground Pipe"
            )
            ax[indexes[i]].plot(
                data_salt_cavern[:, 0] / wind_rating,
                data_salt_cavern[:, 1],
                label="Salt Cavern",
            )
            ax[indexes[i]].plot(
                data_no_storage[:, 0] / wind_rating,
                data_no_storage[:, 1],
                "--k",
                label="No Storage",
            )

            ax[indexes[i]].scatter(
                data_pressure_vessel[np.argmin(data_pressure_vessel[:, 1]), 0]
                / wind_rating,
                np.min(data_pressure_vessel[:, 1]),
                color="k",
            )
            ax[indexes[i]].scatter(
                data_pipe[np.argmin(data_pipe[:, 1]), 0] / wind_rating,
                np.min(data_pipe[:, 1]),
                color="k",
            )
            ax[indexes[i]].scatter(
                data_salt_cavern[np.argmin(data_salt_cavern[:, 1]), 0] / wind_rating,
                np.min(data_salt_cavern[:, 1]),
                color="k",
            )
            ax[indexes[i]].scatter(
                data_no_storage[np.argmin(data_no_storage[:, 1]), 0] / wind_rating,
                np.min(data_no_storage[:, 1]),
                color="k",
                label="Optimal ratio",
            )

            ax[indexes[i]].legend(frameon=False, loc="best")

            ax[indexes[i]].set_xlim([0.2, 2.0])
            ax[indexes[i]].set_ylim([0, 25])

            ax[indexes[i]].annotate("%s MW Wind Plant" % (wind_rating), (0.6, 1.0))

        ax[1, 0].set_xlabel("Electrolyzer/Wind Plant Rating Ratio")
        ax[1, 1].set_xlabel("Electrolyzer/Wind Plant Rating Ratio")
        ax[0, 0].set_ylabel("LCOH ($/kg)")
        ax[1, 0].set_ylabel("LCOH ($/kg)")

        plt.tight_layout()
        plt.savefig(output_dir + "lcoh_vs_rating_ratio.pdf", transparent=True)
        plt.show()

    return 0


def run_policy_options_storage_types(
    verbose=True,
    show_plots=False,
    save_plots=False,
    use_profast=True,
    output_dir="output/",
):

    storage_types = ["pressure_vessel", "pipe", "salt_cavern", "none"]
    policy_options = [1, 2, 3, 4, 5, 6, 7]

    lcoh_array = np.zeros((len(storage_types), len(policy_options)))
    for i, storage_type in enumerate(storage_types):
        for j, poption in enumerate(policy_options):
            lcoh_array[i, j] = run_simulation(
                storage_type=storage_type,
                incentive_option=poption,
                verbose=verbose,
                show_plots=show_plots,
                use_profast=use_profast,
            )
        print(lcoh_array)

    savepath = output_dir + "results/"
    if not os.path.exists(savepath):
        os.makedirs(savepath)
    np.savetxt(
        savepath + "lcoh-with-policy.txt",
        np.c_[np.round(lcoh_array, decimals=2)],
        header="rows: %s, columns: %s"
        % ("".join(storage_types), "".join(str(p) for p in policy_options)),
        fmt="%.2f",
    )

    return 0


def run_policy_storage_design_options(
    verbose=False,
    show_plots=False,
    save_plots=False,
    use_profast=True,
    output_dir="output/",
):

    design_scenarios = [1, 2, 3, 4, 5, 6, 7]
    policy_options = [1, 2, 3, 4, 5, 6, 7]
    storage_types = ["pressure_vessel", "pipe", "salt_cavern", "none"]

    design_series = []
    policy_series = []
    storage_series = []
    lcoh_series = []
    lcoe_series = []
    electrolyzer_capacity_factor_series = []
    annual_energy_breakdown_series = {
        "design": [],
        "policy": [],
        "storage": [],
        "wind_kwh": [],
        "renewable_kwh": [],
        "grid_power_kwh": [],
        "electrolyzer_kwh": [],
        "desal_kwh": [],
        "h2_transport_compressor_power_kwh": [],
        "h2_storage_power_kwh": [],
    }

    lcoh_array = np.zeros((len(design_scenarios), len(policy_options)))
    for i, design in enumerate(design_scenarios):
        for j, policy in enumerate(policy_options):
            for storage in storage_types:
                if storage != "pressure_vessel":  # and storage != "none"):
                    if design != 1 and design != 5 and design != 7:
                        print("skipping: ", design, " ", policy, " ", storage)
                        continue
                design_series.append(design)
                policy_series.append(policy)
                storage_series.append(storage)
                (
                    lcoh,
                    lcoe,
                    capex_breakdown,
                    opex_breakdown_annual,
                    pf_lcoh,
                    electrolyzer_physics_results,
                    pf_lcoe,
                    annual_energy_breakdown,
                ) = run_simulation(
                    storage_type=storage,
                    plant_design_scenario=design,
                    incentive_option=policy,
                    verbose=verbose,
                    show_plots=show_plots,
                    use_profast=use_profast,
                    output_level=3,
                )
                lcoh_series.append(lcoh)
                lcoe_series.append(lcoe)
                electrolyzer_capacity_factor_series.append(
                    electrolyzer_physics_results["H2_Results"]["Life: Capacity Factor"]
                )

                annual_energy_breakdown_series["design"].append(design)
                annual_energy_breakdown_series["policy"].append(policy)
                annual_energy_breakdown_series["storage"].append(storage)
                for key in annual_energy_breakdown.keys():
                    annual_energy_breakdown_series[key].append(
                        annual_energy_breakdown[key]
                    )

    savepath = output_dir + "data/"
    if not os.path.exists(savepath):
        os.makedirs(savepath)
    df = pd.DataFrame.from_dict(
        {
            "Design": design_series,
            "Storage": storage_series,
            "Policy": policy_series,
            "LCOH [$/kg]": lcoh_series,
            "LCOE [$/kWh]": lcoe_series,
            "Electrolyzer capacity factor": electrolyzer_capacity_factor_series,
        }
    )
    df.to_csv(savepath + "design-storage-policy-lcoh.csv")

    df_energy = pd.DataFrame.from_dict(annual_energy_breakdown_series)
    df_energy.to_csv(savepath + "annual_energy_breakdown.csv")
    return 0


def run_design_options(
    verbose=False,
    show_plots=False,
    save_plots=False,
    incentive_option=1,
    output_dir="output/",
):

    design_options = range(1, 8)  # 8
    scenario_lcoh = []
    scenario_lcoe = []
    scenario_capex_breakdown = []
    scenario_opex_breakdown_annual = []
    scenario_pf = []
    scenario_electrolyzer_physics = []

    for design in design_options:
        (
            lcoh,
            lcoe,
            capex_breakdown,
            opex_breakdown_annual,
            pf,
            electrolyzer_physics_results,
        ) = run_simulation(
            verbose=verbose,
            show_plots=show_plots,
            use_profast=True,
            incentive_option=incentive_option,
            plant_design_scenario=design,
            output_level=2,
        )
        scenario_lcoh.append(lcoh)
        scenario_lcoe.append(lcoe)
        scenario_capex_breakdown.append(capex_breakdown)
        scenario_opex_breakdown_annual.append(opex_breakdown_annual)
        scenario_pf.append(pf)
        scenario_electrolyzer_physics.append(electrolyzer_physics_results)
    df_aggregate = pd.DataFrame.from_dict(
        {
            "Design": [int(x) for x in design_options],
            "LCOH [$/kg]": scenario_lcoh,
            "LCOE [$/kWh]": scenario_lcoe,
        }
    )
    df_capex = pd.DataFrame(scenario_capex_breakdown)
    df_opex = pd.DataFrame(scenario_opex_breakdown_annual)

    df_capex.insert(0, "Design", design_options)
    df_opex.insert(0, "Design", design_options)

    # df_aggregate = df_aggregate.transpose()
    df_capex = df_capex.transpose()
    df_opex = df_opex.transpose()

    results_path = output_dir + "/combined_results/"
    if not os.path.exists(results_path):
        os.mkdir(results_path)
    df_aggregate.to_csv(results_path + "metrics.csv")
    df_capex.to_csv(results_path + "capex.csv")
    df_opex.to_csv(results_path + "opex.csv")
    return 0


def run_storage_options(output_dir="output/"):
    storage_types = ["pressure_vessel", "pipe", "salt_cavern", "none"]
    lcoe_list = []
    lcoh_list = []
    lcoh_with_grid_list = []
    lcoh_grid_only_list = []
    for storage_type in storage_types:
        lcoe, lcoh, _ = run_simulation(
            verbose=False,
            show_plots=False,
            save_plots=False,
            use_profast=True,
            incentive_option=1,
            plant_design_scenario=1,
            storage_type=storage_type,
            output_level=4,
            grid_connection=False,
            output_dir=output_dir,
        )
        lcoe_list.append(lcoe)
        lcoh_list.append(lcoh)

        # with grid
        _, lcoh_with_grid, lcoh_grid_only = run_simulation(
            verbose=False,
            show_plots=False,
            save_plots=False,
            use_profast=True,
            incentive_option=1,
            plant_design_scenario=1,
            storage_type=storage_type,
            output_level=4,
            grid_connection=True,
            output_dir=output_dir,
        )
        lcoh_with_grid_list.append(lcoh_with_grid)
        lcoh_grid_only_list.append(lcoh_grid_only)

    data_dict = {
        "Storage Type": storage_types,
        "LCOE [$/MW]": np.asarray(lcoe_list) * 1e3,
        "LCOH [$/kg]": lcoh_list,
        "LCOH with Grid [$/kg]": lcoh_with_grid_list,
        "LCOH Grid Only [$/kg]": lcoh_grid_only_list,
    }
    df = pd.DataFrame.from_dict(data_dict)

    savepath = output_dir + "data/"
    if not os.path.exists(savepath):
        os.makedirs(savepath)
    df.to_csv(savepath + "storage-types-and-matrics.csv")
    return 0
